"""
ACR Low-Contrast Object Detectability Task
__________________________________________

Reference
_________

`ACR Large Phantom Guidance PDF <https://accreditationsupport.acr.org/helpdesk/attachments/11093487417>`_

Intro
_____

Performs a series of Image Processing steps that attempts to

    #. Isolate the signal of the spots inside the circle ROI in slices 8 through 11.
    #. Predict where spots should be present in ROI.
    #. Walk through the predicted spot ROIs.
    #. Asses if spoke is valid.
    #. Stop at first invalid spoke.
    #. Calculate the slice score.

A lot of the tuning performed attempts to strike a balance between the real number of high intensity signals and
avoiding overestimation. As a result, my current attempt is biased towards underestimation if the signal is too
small to survive thresholding.

ACR Guidelines
______________

ACR Algorithm
+++++++++++++

    #. Display slice 11, which has the highest contrast objects. Adjust the display window width and level
        settings for best visibility of the low contrast objects. This will usually require a fairly narrow window
        width and careful adjustment of the level to best distinguish the objects from the background.
    #. Count the number of complete spokes. Begin counting with the spoke having the largest diameter
        disks; this spoke is at 12 o clock or slightly to the right of 12 o clock, and is referred to as spoke 1.
        Count clockwise from spoke 1 until a spoke is reached where one or more of the disks is not
        discernible from the background. A spoke is complete only if all three disks are discernible. Count
        complete spokes, not individual disks.
    #. The score for this slice is the number of complete spokes. Record the score.
    #. Repeat the procedure to determine the number of visible spokes for the remaining LCD images.

ACR Scoring Rubric
++++++++++++++++++

For each series, record the number of complete spokes visible on each slice, then sum the values for all four
slices to determine the total LCD score. For example, if the ACR T2 series scored 3 spokes in slice 8, 5 spokes
in slice 9, 9 spokes in slice 10, and 10 spokes in slice 11; the total score for the ACR T2 series would be 3 + 5
+ 9 + 10 = 27.

Nominal Field  ACR T1 LCD       ACR T2 LCD
Strength       Limit           Limit
               (total spokes)  (total spokes)
_____________  _______________ ______________
<1.5T           7               7
1.5T - <3T      30              25
3T              37              37

Notes
_____

The approach taken here was to do a Difference of Gaussians with several blurring and thresholding steps such that
the resulting image has most of the noise dropped. Then, we dilate whatever signal is present and hope that the predicted
mask sees the signal. This approach is very error prone with very noisy scans. As a result, we threshold such that we
only consider a very high percentile threshold point based on the available signal. Furthermore, the DoG operation tends
to introduce a rim of noise on the periphery which often blends with the outer ring of spots. Per my tuning, this is
minimized, but still exists. To avoid this artifact, I can decrease the aggressiveness of the DoG operation at the cost
of more residual noise in the final output. The reason I tried this approach first is that I was interested in
avoiding adding more dependencies and training a neural network for this task at this moment. However, current testing
suggests that I should read about blob detection algorithms and consider machine learning as well. It could be a very
productive approach long term.

The Hough Transform does not work reliably for this task. You really are looking into blob detection more so than circle
detection. DoG returns a decently cleaned image for further analysis. The result is just not easy to clean via
thresholding.

TL;DR; I definitely need to return to this task and try a cleaner approach using blob detection.

Created by Luis M. Santos, M.D.
luis.santos2@nih.gov

12/02/2025
"""

# Python imports
import os

# Module imports
import cv2
import numpy as np
from hazenlib import logger
from hazenlib.ACRObject import ACRObject
from hazenlib.HazenTask import HazenTask
from hazenlib.utils import (
    compute_radius_from_area,
    create_circular_roi_at,
    expand_data_range,
    wait_on_parallel_results,
)
from hazenlib.types import Measurement
from matplotlib.pyplot import subplots as plt_subplots


class ACRObjectDetectability(HazenTask):
    """Uniformity measurement class for DICOM images of the ACR phantom."""

    DOT_SEPARATION = 12.8  #: 12 to 14 mm from center to center of each circle.
    DOT_ANGLE = np.deg2rad(36)  #: Each spoke is at this angle of separation.
    SLICE_ANGLE_OFFSET = np.deg2rad(9)  #: Each subsequent slice
    START_ANGLE = np.deg2rad(90)  #: Slice 0 has spots at a 90 deg angle
    ORIG_SPOKE_RADII = {
        0: 3.5,
        1: 3.1945,
        2: 2.889,
        3: 2.5835,
        4: 2.278,
        5: 1.9725,
        6: 1.667,
        7: 1.3615,
        8: 1.056,
        9: 0.75,
    }
    SPOKE_RADII = {  #: Radius used for each spoke spot. Meaning, in spoke 0 all spots have the same
        0: 2.0,  #: radius.
        1: 2.0,
        2: 2.0,
        3: 2.0,
        4: 2.0,
        5: 2.0,
        6: 2.0,
        7: 2.0,
        8: 2.0,
        9: 2.0,
    }
    BINARIZATION_THRESHOLD = {1.5: 97.8, 3.0: 97.5}
    FIRST_SLICE_NUM = 8

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # Initialise ACR object
        self.ACR_obj = ACRObject(self.dcm_list)
        # Required pixel radius to produce ~75cm2 ROI
        self.r_inner = compute_radius_from_area(80, self.ACR_obj.dx)
        # Required pixel radius to produce ~0.25cm2 ROI
        self.r_binarization_sample = compute_radius_from_area(
            45, self.ACR_obj.dx
        )
        # Required pixel radius to produce ~15cm2 ROI
        self.r_noise = compute_radius_from_area(45, self.ACR_obj.dx)
        # Required pixel radius to produce ~0.25cm2 ROI
        self.r_spot = compute_radius_from_area(0.25, self.ACR_obj.dx)

    def run(self) -> dict:
        """Main function for performing uniformity measurement using slice 7 from the ACR phantom image set.

        Returns:
            dict: results are returned in a standardised dictionary structure specifying the task name, input DICOM
                Series Description + SeriesNumber + InstanceNumber, task measurement key-value pairs, optionally path
                to the generated images for visualisation
        """
        slices = [
            self.ACR_obj.slice_stack[7],
            self.ACR_obj.slice_stack[8],
            self.ACR_obj.slice_stack[9],
            self.ACR_obj.slice_stack[10],
        ]
        # Initialise results dictionary
        results = self.init_result_dict()
        results.files = tuple([self.img_desc(sl) for sl in slices])

        r = self.get_spokes_and_scores(slices)

        if not len(r["meta"]["measurement"]):
            msg = "No low contrast object detection measurements made"
            logger.error(msg)
            raise ValueError(msg)

        for measurement, score in r["meta"]["measurement"].items():
            subtype = (
                f"slice {measurement}"
                if isinstance(measurement, int)
                else "total"
            )

            m = Measurement(
                name="LowContrastObjectDetectability",
                value=score,
                type="measured",
                subtype=subtype,
                description=(
                    f"Field Strength = {r['meta']['field_strength']}T"
                ),
            )
            results.add_measurement(m)

        # only return reports if requested
        if self.report:
            results.add_report_image(self.report_files)

        return results

    def write_report(self, slices, results):
        data = results["data"]

        arg_list = [
            (slices[i], data[i], data[i]["center"]) for i in range(len(slices))
        ]
        self.report_files = wait_on_parallel_results(
            self.write_report_slice, arg_list
        )

    def write_report_slice(
        self, dcm, img_result, center, theta=np.linspace(0, 2 * np.pi, 360)
    ):
        (center_x, center_y) = center
        fig, axes = plt_subplots(4, 1)
        fig.set_size_inches(8, 16)
        fig.tight_layout(pad=4)

        # Centroid
        axes[0].imshow(img_result["img"][0], cmap="gray", vmin=0, vmax=255)
        axes[0].scatter(center_x, center_y, c="red")
        axes[0].axis("off")
        axes[0].set_title("Window Leveled + Centroid Location")

        # DoG
        axes[1].imshow(img_result["img"][1], cmap="viridis")
        axes[1].axis("off")
        axes[1].set_title("Difference of Gaussians")

        # Dilated
        axes[2].imshow(img_result["img"][2], cmap="viridis")
        axes[2].axis("off")
        axes[2].set_title("Filtered (binarized + dilated)")

        # axes[2].imshow(img_result['img'][2], cmap='gray', vmin=0, vmax=255)
        axes[3].imshow(img_result["img"][3], cmap="viridis")
        for spoke in img_result["spokes"]:
            spot_center1, spot_center2, spot_center3 = spoke["centers"]
            axes[3].plot(
                self.r_spot * np.cos(theta) + spot_center1[0],
                self.r_spot * np.sin(theta) + spot_center1[1],
                c="green",
            )
            axes[3].plot(
                self.r_spot * np.cos(theta) + spot_center2[0],
                self.r_spot * np.sin(theta) + spot_center2[1],
                c="green",
            )
            axes[3].plot(
                self.r_spot * np.cos(theta) + spot_center3[0],
                self.r_spot * np.sin(theta) + spot_center3[1],
                c="green",
            )
        axes[3].axis("off")
        axes[3].set_title("Valid Spokes (dilated)")

        img_path = os.path.realpath(
            os.path.join(self.report_path, f"{self.img_desc(dcm)}.png")
        )
        fig.savefig(img_path)
        return img_path

    @staticmethod
    def calculate_spot_location(center, angle, spot_separation, spot):
        x_dist, y_dist = (
            np.floor(np.cos(angle) * spot_separation * spot),
            np.floor(np.sin(angle) * spot_separation * spot),
        )
        return (center[0] + x_dist, center[1] - y_dist)

    def detect_spot(
        self, img, center, angle, spot_separation, spot_radius, spot
    ):
        x, y = self.calculate_spot_location(
            center, angle, spot_separation, spot
        )
        return create_circular_roi_at(img, spot_radius, x, y), (x, y)

    def detect_spoke(self, img, center, slice_num, spoke):
        spot_radius = int(np.ceil(self.SPOKE_RADII[spoke] / self.ACR_obj.dx))
        spot_separation = self.DOT_SEPARATION / self.ACR_obj.dx
        angle = (
            self.START_ANGLE
            - (spoke * self.DOT_ANGLE)
            - (self.SLICE_ANGLE_OFFSET * slice_num)
        )

        # Generate individual spot masks on image.
        spot1 = self.detect_spot(
            img, center, angle, spot_separation, spot_radius, 1
        )
        spot2 = self.detect_spot(
            img, center, angle, spot_separation, spot_radius, 2
        )
        spot3 = self.detect_spot(
            img, center, angle, spot_separation, spot_radius, 3
        )

        return spot1, spot2, spot3

    @staticmethod
    def combine_masks(spots, target_mask):
        # Combine the spot masks into a master spoke mask
        spot1, spot2, spot3 = spots
        combined_mask = np.ma.mask_or(
            ~spot1[0].mask, np.ma.mask_or(~spot2[0].mask, ~spot3[0].mask)
        )
        return np.ma.mask_or(combined_mask, target_mask)

    def compute_score(self, feature_data, center, slice_num):
        spoke_results = {}
        for spoke in range(10):
            spot1, spot2, spot3 = self.detect_spoke(
                feature_data, center, slice_num, spoke
            )
            spot_score = sum(
                [spot1[0].sum() > 0, spot2[0].sum() > 0, spot3[0].sum() > 0]
            )
            valid = spot_score == 3
            if valid:
                spoke_results[spoke] = {
                    "spots": spot_score,
                    "centers": [spot1[1], spot2[1], spot3[1]],
                }
            else:
                break
        return spoke_results

    def get_img_center(self, img):
        (center_x, center_y), _ = self.ACR_obj.find_phantom_center(
            img, self.ACR_obj.dx, self.ACR_obj.dy
        )
        offsetted_y = np.round(center_y + 5 / self.ACR_obj.dy)
        return (np.round(center_x - self.ACR_obj.dx), np.round(offsetted_y))

    def detect_objects(self, dcm, field_strength, slice_num):
        """Performs all preprocessing steps required to isolate the signal from background. Then, runs an ROI based
        algorithm to detect the presence of valid spokes in dataset.

        Steps
        _____

            #. Request presentation level pixel data.
            #. Finds centroid of data.
            #. Restrict workspace to the inner ROI in dataset.
            #. Computes Histogram based Window Center and Window Width
            #. Performs a small Gaussian pass to denoise the input image.
            #. Performs Difference of Gaussians (DoG) pass.
            #. Performs another small Gaussian filter pass to decimate some of the remaining noise.
            #. Performs binarization of dog based on a percentile.
            #. Dilate remaining signal since we have destroyed so much of it by this point.
            #. Detect and count spots/spokes to generate score.
            #. Return scores and relevant images.

        Notes
        _____

        ..note::

            The weird Gaussian filtering passes not associated with the DoG have subtle and profound effects on what
            signal survives all of our filtering levers.

        ..note::

            The DoG implementation used here also allows for the application of gamma correction and multiple passes.
            However, none of these options ended up being necessary on the final version of this algorithm. I made
            the implementations available as individual methods in the ACRObject class for other future image
            manipulation tasks.

        ..warning::

            The scoring algorithm is a very dumb algorithm in that it computes the center of a spot ROI and performs
            a sum on the captured pixel population. Any values > 0 are assumed to imply that the neighborhood signal
            must have come from a strong signal. However, this is only an approximation to reality and thus it is
            weak against inputs that are noisy after all of our filtering steps. As a result, it is strongly encouraged
            to provide a copy of the dilated output in the report to allow for visual inspection of the noise
            distribution. I am not clever enough to figure out how to detect this kind of noise after generating
            results.

        ..alert::

            The GE dataset present in this repository is particularly troublesome because it represents a case in which
            there is a poor SNR on a high resolution (0.5mm) 512x512 FoV. This is a problem because the ACR prescribed
            that the matrix has to be 256x256 with a resolution of 1.0mm. The main issue is if I make the algorithm
            robust to the ACR Phantom requirements, the algorithm becomes weak to the GE data's 8th slice, which is
            pure noise. The algorithm somehow detects the alignment of the noise spots and yields at least 2 valid
            spokes when that should not be the case.

        Args:
            dcm (pydicom.Dataset): DICOM image object to calculate uniformity from.
            field_strength (float): Scanner field strength parameter. Kinda needed to get the correct level of noise
                cleaning.
            slice_num (int): Index of this DICOM slice in the original sorted dicom stack.

        Returns:
            dict: Relevant results that can be reported on. These include the per slice score and intermediate images.

            .. code-block::
                :caption: Results Structure!

                {
                    'id': slice_id,
                    'img': [windowed, dog, dilated, dilated],
                    'spokes': [spoke for spoke in results.values()],
                    'score': len(results),
                    'center': (center_x, center_y)
                }
        """
        slice_id = self.FIRST_SLICE_NUM + slice_num
        logger.info(f"Processing slice # {slice_id}")

        img, rescaled, presentation = self.ACR_obj.get_presentation_pixels(dcm)

        # Step 0, let's obtain a better center
        (center_x, center_y) = self.get_img_center(rescaled)
        logger.info(
            f"Phantom centroid set to {(center_x, center_y)} for slice {slice_id}!"
        )

        # Now, we can ready the ROI on which to focus on extracting the signal
        inner_roi = create_circular_roi_at(
            rescaled, self.r_inner, center_x, center_y
        )
        inner_roi[inner_roi.mask] = 0

        # Find noise sampling ROI
        noise_roi = create_circular_roi_at(
            inner_roi, self.r_noise, center_x, center_y
        )
        noise_roi[noise_roi.mask] = 0

        # Compute the window level and width in the noise sampling area.
        center, width = self.ACR_obj.compute_center_and_width(noise_roi)
        logger.info(f"Target Windowing => {center}, {width}")

        # Apply the previously computed window settings to the inner ROI window.
        # We adjust the windowing here using the clip method which is a custom method. The clip method is not one
        # of the standard DICOM windowing methods. It is a modified linear exact method except we do not rescale the
        # window data. Rescaling the window data yields promotion of noise to the same rank as our signal.
        contrasted = self.ACR_obj.apply_window_center_width(
            inner_roi, center, width * field_strength
        )

        # First, let do a light Gaussian pass to help remove some of the crazy noise and improve SNR.
        # Now, testing against the GE test dataset with a 512x512 matrix from a 1.5T scanner, my algorithm favors a sigma
        # of 1 or higher. Testing against high resolution Philips 3T dataset favors a fractional sigma <= 0.5.
        # The issue is that the GE dataset represents a noisy 1.5T scan acquisition with enough SNR for a human
        # to make a judgement call biased towards overestimating spoke count. This same noise causes a small
        # overestimation from the algorithm on the worst case slice due to two areas of high SNR. However, that GE slice
        # should have yielded 0 valid spokes. As a result, the compromise seems to lie somewhere between a 0.3 to 0.5
        # factor which adjusted by the pixel resolution will yield a sigma > 0.5 for the high resolution GE and
        # 0.3 for the lower (1mm) resolution acquisitions.
        resolution_factor = 1 / self.ACR_obj.dx
        noise_removed = self.ACR_obj.filter_with_gaussian(
            contrasted, 0.7 * resolution_factor
        )
        noise_removed = np.ma.masked_array(
            noise_removed, mask=inner_roi.mask, fill_value=0
        )

        # Perform Difference of Gaussians to further isolate relevant pixels
        # Using a large gamma to allow high intensities to survive the DoG operation.
        # Gamma correction can have a profound effect on remaining signal just like the selection of the sigmas.
        # Gamma correction here helps with fine-tuning to approximate what I thought is reality for the GE dataset which
        # was noisier than more ideal scans. GE => gamma = 20, sigma2 = 3.5
        factor = 3 / field_strength
        dog = self.ACR_obj.filter_with_dog(
            noise_removed,
            (0.1 * factor) / self.ACR_obj.dx,
            (1.5 * factor) / self.ACR_obj.dx,
        )
        dog = self.ACR_obj.filter_with_gaussian(dog, 0.5 / self.ACR_obj.dx)
        dog = np.ma.masked_array(dog, mask=inner_roi.mask, fill_value=0)

        # Binarize the results
        # Results should be so clean at this stage that we do not need to be aggressive (97+) in the thresholding.
        # 92% is pretty safe and accurate in most cases, but the 1.5T can yield more human like results with a slightly
        # lesser percentile. I found that 91st percentile strikes an acceptable balance between surviving noise and
        # keeping the dimmer spots in the signal population.
        binarized = self.ACR_obj.binarize_image(
            dog.copy(), self.BINARIZATION_THRESHOLD.get(field_strength, 92)
        )

        # Dilate the signal that is present.
        dilated = cv2.dilate(
            binarized, cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (3, 3))
        )
        dilated = np.ma.masked_array(
            dilated, mask=inner_roi.mask, fill_value=0
        )

        # Count spots and spokes clockwise
        # A spot is valid if its max intensity is above relative threshold
        # A spoke is valid if it contains 3 successive spots in diagonal.
        results = self.compute_score(dilated, (center_x, center_y), slice_num)

        windowed = expand_data_range(contrasted)
        windowed[inner_roi.mask] = 0

        return {
            "id": slice_id,
            "img": [windowed, dog, dilated, dilated],
            "spokes": [spoke for spoke in results.values()],
            "score": len(results),
            "center": (center_x, center_y),
        }

    def get_spokes_and_scores(self, slices):
        """For each slice, we go through a series of image filtering steps. Once the image is sufficiently filtered,
        we then predict where spot ROIs are expected to be. If the ROI captures a population of intensities > 0, the
        spot is considered valid. Three valid spots in a radial pattern yields a valid spoke. Scoring stops at the
        first incomplete spoke. We then collect the per slice and overall scores as well as

        Args:
            slices (list of pydicom.Dataset): list of relevant slices (8 to 11) used for generating ACR scores.

        Returns:
            dict: Results including individual slice scores, images, and scanner information.

            .. code-block::
                :caption: Results Structure!

                {
                    "meta": {
                        "field_strength": field_strength,
                        "slice_scores": [],
                        "score": 0
                    },
                    "data": {
                        0: {slice results},
                        1: {slice results},
                        2: {slice results},
                        3: {slice results},
                    }
                }
        """
        field_strength = slices[-1].MagneticFieldStrength

        results = {
            "meta": {"field_strength": field_strength, "measurement": {}},
            "data": {},
        }

        # Run processing jobs
        jobs = [(slices[i], field_strength, i) for i in range(4)]
        result_data = wait_on_parallel_results(self.detect_objects, jobs)

        # Collect data and final score
        score = 0
        for i in range(4):
            results["data"][i] = result_data[i]
            slice_score = results["data"][i]["score"]
            results["meta"]["measurement"][self.FIRST_SLICE_NUM + i] = (
                slice_score
            )
            score += slice_score

        # Append meta data about results
        results["meta"]["measurement"]["total_score"] = score

        # Generate report
        if self.report:
            logger.info("Writing report ... ")
            self.write_report(slices, results)
            logger.info("Finished writing report!")

        return results



"""
SNR(Im)

Calculates the SNR for a single-slice image of a uniform MRI phantom

This script utilises the smoothed subtraction method described in McCann 2013:
A quick and robust method for measurement of signal-to-noise ratio in MRI, Phys. Med. Biol. 58 (2013) 3775:3790


Created by Neil Heraghty

04/05/2018
"""

from __future__ import annotations

# Python imports
import contextlib
import os
from typing import Any

# Module imports
import cv2 as cv
import numpy as np
import pydicom
import skimage.filters
from pathlib import Path
from scipy import ndimage

# Local imports
import hazenlib.exceptions as exc
import hazenlib.utils
from hazenlib.HazenTask import HazenTask
from hazenlib.logger import logger
from hazenlib.types import Measurement, Result
from hazenlib.utils import dcmread


class SNR(HazenTask):
    """Signal-to-noise ratio measurement class for DICOM images of the MagNet phantom

    Inherits from HazenTask class
    """

    def __init__(
        self,
        measured_slice_width: float | None = None,
        coil: str | None = None,
        subtract: str | Path | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialise the Hazen ACR SNR Object."""
        super().__init__(**kwargs)

        # Pulls out the subtract path (for compatability with acr_snr)
        if subtract is not None:
            dicom_files = list(Path(subtract).glob("*.dcm"))
            if len(dicom_files) > 1:
                logger.warning(
                    "Multiple DICOM files found taking the first: %s",
                    dicom_files[0],
                )

            if len(dicom_files) < 1:
                logger.warning(
                    "No DICOM files found in: %s - ignore subtraction.",
                    subtract,
                )
            else:
                self.dcm_list.append(dcmread(dicom_files[0]))

        # measured slice width is expected to be a floating point number
        self.measured_slice_width = measured_slice_width
        with contextlib.suppress(TypeError):
            self.measured_slice_width = float(measured_slice_width)

        # Determining kernel size based on coil choice. Values of 9 and 25 come from McCann 2013 paper.
        try:
            if coil.lower() in ["hc", "head"]:
                self.kernel_size = 9
            elif coil.lower() in ["bc", "body"]:
                self.kernel_size = 25

        except AttributeError:
            self.kernel_size = 9

    def run(self) -> Result:
        """Main function for performing signal-to-noise ratio measurement

        Notes:
            Five square ROIs are created, one at the image centre, and four peripheral
            ROIs with their centres displaced at 45, 135, 225 and 315 degrees from the
            centre.

        Returns:
            dict: results are returned in a standardised dictionary structure specifying the task name, input DICOM Series Description + SeriesNumber + InstanceNumber, task measurement key-value pairs, optionally path to the generated images for visualisation
        """
        results = self.init_result_dict()
        results.files = [self.img_desc(img) for img in self.dcm_list]

        # SUBTRACTION METHOD with a pair of input files
        if len(self.dcm_list) == 2:
            snr, normalised_snr = self.snr_by_subtraction(
                self.dcm_list[0], self.dcm_list[1], self.measured_slice_width
            )
            results.add_measurement(
                Measurement(
                    name="SNR",
                    subtype="subtraction",
                    type="measured",
                    value=round(snr, 2),
                    visibility="intermediate",
                ),
            )
            results.add_measurement(
                Measurement(
                    name="SNR",
                    subtype="subtraction",
                    type="normalised",
                    value=round(normalised_snr, 2),
                ),
            )

        # SINGLE METHOD (SMOOTHING) for every input file
        for dcm in self.dcm_list:
            snr, normalised_snr = self.snr_by_smoothing(
                dcm, self.measured_slice_width
            )
            results.add_measurement(
                Measurement(
                    name="SNR",
                    subtype="smoothing",
                    type="measured",
                    description=self.img_desc(dcm),
                    value=round(snr, 2),
                    visibility="intermediate",
                ),
            )
            results.add_measurement(
                Measurement(
                    name="SNR",
                    subtype="smoothing",
                    type="normalised",
                    description=self.img_desc(dcm),
                    value=round(normalised_snr, 2),
                ),
            )

        # only return reports if requested
        if self.report:
            results.add_report_image(self.report_files)

        return results

    def two_inputs_match(
        self, dcm1: pydicom.Dataset, dcm2: pydicom.Dataset
    ) -> bool:
        """Check if two DICOMs are sufficiently similar, based on the following fields
            "StudyInstanceUID", "RepetitionTime", "EchoTime", "FlipAngle"

        Args:
            dcm1 (pydicom.Dataset): DICOM object to compare
            dcm2 (pydicom.Dataset): _description_

        Returns:
            bool: _description_
        """

        fields_to_match = [
            "StudyInstanceUID",
            "RepetitionTime",
            "EchoTime",
            "FlipAngle",
        ]

        for field in fields_to_match:
            if dcm1.get(field) != dcm2.get(field):
                return False
        return True

    def get_normalised_snr_factor(
        self, dcm: pydicom.Dataset, measured_slice_width=None
    ) -> float:
        """Calculates SNR normalisation factor.

        Notes:
            Method matches MATLAB script.
            Utilises user provided slice_width if provided. Else finds from dcm.
            Finds dx, dy and bandwidth from dcm.
            Seeks to find TR, image columns and rows from dcm. Else uses default values.

        Args:
            dcm (pydicom.Dataset): DICOM image object
            measured_slice_width (float or None): slice width from user input

        Returns:
            float: normalised snr factor
        """

        dx, dy = hazenlib.utils.get_pixel_size(dcm)
        bandwidth = hazenlib.utils.get_bandwidth(dcm)
        TR = hazenlib.utils.get_TR(dcm)
        rows = hazenlib.utils.get_rows(dcm)
        columns = hazenlib.utils.get_columns(dcm)

        if measured_slice_width:
            slice_thickness = measured_slice_width
        else:
            slice_thickness = hazenlib.utils.get_slice_thickness(dcm)

        averages = hazenlib.utils.get_average(dcm)
        bandwidth_factor = np.sqrt((bandwidth * columns / 2) / 1000) / np.sqrt(
            30
        )
        voxel_factor = 1 / (0.001 * dx * dy * slice_thickness)

        normalised_snr_factor = (
            bandwidth_factor
            * voxel_factor
            * (1 / (np.sqrt(averages * rows * (TR / 1000))))
        )

        return normalised_snr_factor

    def filtered_image(self, dcm: pydicom.Dataset) -> np.array:
        """Performs a 2D convolution (for filtering images)
        using uniform_filter SciPy function and a kernel size based on user input coil

        Args:
            dcm (pydicom.Dataset): DICOM image to be filtered

        Returns:
            np.array: filtered image pixel values
        """
        a = dcm.pixel_array.astype("int")

        # filter size = 9, following MATLAB code and McCann 2013 paper for head coil, although note McCann 2013 recommends 25x25 for body coil.

        # 9 for head coil, 25 for body coil
        filtered_array = ndimage.uniform_filter(
            a, self.kernel_size, mode="constant"
        )
        return filtered_array

    def get_noise_image(self, dcm: pydicom.Dataset) -> np.array:
        """Separates the image noise
        by smoothing the image and subtracting the smoothed image from the original.

        Args:
            dcm (pydicom.Dataset): DICOM image to get noise from

        Returns:
            np.array: pixel array representing the image noise
        """
        a = dcm.pixel_array.astype("int")

        # Convolve image with boxcar/uniform kernel
        imsmoothed = self.filtered_image(dcm)

        # Subtract smoothed array from original
        imnoise = a - imsmoothed

        return imnoise

    def threshold_image(self, dcm: pydicom.Dataset):
        """Determine threshold and mask based on image

        Args:
            dcm (pydicom.Dataset): DICOM image to get noise from

        Returns:
            tuple of np.array: (imthresholded, mask)
                pixel array representing the image above threshold
                and a corresponding mask
        """
        a = dcm.pixel_array.astype("int")

        # threshold_li: Pixels > this value are assumed foreground
        threshold_value = skimage.filters.threshold_li(a)
        # print('threshold_value =', threshold_value)
        mask = a > threshold_value
        imthresholded = np.zeros_like(a)
        imthresholded[mask] = a[mask]

        # # For debugging: Threshold figures:
        # from matplotlib import pyplot as plt
        # plt.figure()
        # fig, ax = plt.subplots(2, 2)
        # ax[0, 0].imshow(a)
        # ax[0, 1].imshow(mask)
        # ax[1, 0].imshow(imthresholded)
        # ax[1, 1].imshow(a-imthresholded)
        # fig.savefig("../THRESHOLD.png")

        return imthresholded, mask

    def get_binary_mask_centre(self, binary_mask) -> (int, int):
        """Determine coordinates of binary polygonal shape's centre

        Args:
            binary_mask: mask of a shape

        Returns:
            tuple of int corresponding to centroid_coords: (col:int, row:int)
        """

        from skimage import util
        from skimage.measure import label, regionprops

        img = util.img_as_ubyte(binary_mask) > 0
        label_img = label(img, connectivity=img.ndim)
        props = regionprops(label_img)
        col = int(props[0].centroid[0])
        row = int(props[0].centroid[1])
        # print('Centroid coords [x,y] =', col, row)

        return int(col), int(row)

    def get_roi_samples(
        self,
        ax,
        dcm: pydicom.Dataset or np.ndarray,
        centre_col: int,
        centre_row: int,
    ) -> list:
        """Determine region of interest from a pixel array.

        Args:
            ax (matplotlib axes): diagram axis for visualisation with matplotlib
            dcm (pydicom.Dataset or np.ndarray): image pixel array
            centre_col (int): center coordinate column
            centre_row (int): center coordinate row

        Returns:
            list of np.ndarray: corresponding to pixel array subsets at predefined ROI

        """
        data = dcm if isinstance(dcm, np.ndarray) else dcm.pixel_array

        sample = [None] * 5
        # for array indexing: [row, column] format
        sample[0] = data[
            (centre_row - 10) : (centre_row + 10),
            (centre_col - 10) : (centre_col + 10),
        ]
        sample[1] = data[
            (centre_row - 50) : (centre_row - 30),
            (centre_col - 50) : (centre_col - 30),
        ]
        sample[2] = data[
            (centre_row + 30) : (centre_row + 50),
            (centre_col - 50) : (centre_col - 30),
        ]
        sample[3] = data[
            (centre_row - 50) : (centre_row - 30),
            (centre_col + 30) : (centre_col + 50),
        ]
        sample[4] = data[
            (centre_row + 30) : (centre_row + 50),
            (centre_col + 30) : (centre_col + 50),
        ]

        if ax:
            from matplotlib.collections import PatchCollection
            from matplotlib.patches import Rectangle

            # for patches: [column/x, row/y] format

            rects = [
                Rectangle((centre_col - 10, centre_row - 10), 20, 20),
                Rectangle((centre_col - 50, centre_row - 50), 20, 20),
                Rectangle((centre_col + 30, centre_row - 50), 20, 20),
                Rectangle((centre_col - 50, centre_row + 30), 20, 20),
                Rectangle((centre_col + 30, centre_row + 30), 20, 20),
            ]
            pc = PatchCollection(
                rects, edgecolors="red", facecolors="None", label="ROIs"
            )
            ax.add_collection(pc)

        return sample

    def get_object_centre(self, dcm) -> (int, int):
        """Find the phantom object within the image and return its centre col and row value.
        Note first element in output = col, second = row.

        Args:
            dcm (pydicom.Dataset): DICOM image to get noise from

        Returns:
            tuple of int corresponding to centroid_coords: (col:int, row:int)
        """

        # Shape Detection
        try:
            logger.debug("Performing phantom shape detection.")
            shape_detector = hazenlib.utils.ShapeDetector(arr=dcm.pixel_array)
            orientation = hazenlib.utils.get_image_orientation(dcm)

            if orientation in ["Sagittal", "Coronal"]:
                logger.debug("Orientation = sagittal or coronal.")
                # orientation is sagittal to patient
                try:
                    (col, row), size, angle = shape_detector.get_shape(
                        "rectangle"
                    )
                except exc.ShapeError as e:
                    # shape_detector.find_contours()
                    # shape_detector.detect()
                    # contour = shape_detector.shapes['rectangle'][1]
                    # angle, centre, size = cv.minAreaRect(contour)
                    # print((angle, centre, size))
                    # im = cv.drawContours(dcm.pixel_array.copy(), [shape_detector.contours[0]], -1, (0, 255, 255), 10)
                    # plt.imshow(im)
                    # plt.savefig("rectangles.png")
                    # print(shape_detector.shapes.keys())
                    raise e
            elif orientation == "Transverse":
                logger.debug("Orientation = transverse.")
                try:
                    col, row, r = shape_detector.get_shape("circle")
                except exc.MultipleShapesError:
                    logger.info(
                        "Warning! Found multiple circles in image, will assume largest circle is phantom."
                    )
                    col, row, r = self.get_largest_circle(
                        shape_detector.shapes["circle"]
                    )
            else:
                raise exc.ShapeError("Unable to identify phantom shape.")

        # Threshold Detection
        except exc.ShapeError:
            logger.info(
                "Shape detection failed. Performing object centre measurement by thresholding."
            )
            _, mask = self.threshold_image(dcm)
            row, col = self.get_binary_mask_centre(mask)

        return int(col), int(row)

    def snr_by_smoothing(
        self, dcm: pydicom.Dataset, measured_slice_width=None
    ):
        """Calculate signal to noise ratio by smoothing

        Args:
            dcm (pydicom.Dataset): DICOM image object
            measured_slice_width (float or None): slice width from user input

        Returns:
            tuple of float: SNR and normalised SNR values

        """
        col, row = self.get_object_centre(dcm)
        noise_img = self.get_noise_image(dcm)

        signal = [
            np.mean(roi)
            for roi in self.get_roi_samples(
                ax=None, dcm=dcm, centre_col=col, centre_row=row
            )
        ]

        noise = [
            np.std(roi, ddof=1)
            for roi in self.get_roi_samples(
                ax=None, dcm=noise_img, centre_col=col, centre_row=row
            )
        ]
        # note no root_2 factor in noise for smoothed subtraction (one image) method, replicating Matlab approach and McCann 2013

        snr = np.mean(np.divide(signal, noise))

        normalised_snr = snr * self.get_normalised_snr_factor(
            dcm, measured_slice_width
        )

        if self.report:
            import matplotlib.pyplot as plt

            fig, axes = plt.subplots(1, 1)
            fig.set_size_inches(5, 5)
            fig.tight_layout(pad=1)

            axes.set_title("smoothed noise image")
            axes.imshow(noise_img, cmap="gray", label="smoothed noise image")
            axes.scatter(col, row, 10, marker="+", label="centre")
            self.get_roi_samples(axes, dcm, col, row)
            axes.legend()

            img_path = os.path.realpath(
                os.path.join(
                    self.report_path, f"{self.img_desc(dcm)}_smoothing.png"
                )
            )
            fig.savefig(img_path)
            self.report_files.append(img_path)

        return snr, normalised_snr

    def get_largest_circle(self, circles):
        """Determine circle with largest radius from list of detected circles

        Args:
            circles (_type_): _description_

        Returns:
            tuple: of centre coordinates (col, row) and radius (float)
        """
        largest_col, largest_row, largest_r = 0, 0, 0
        for circle in circles:
            (col, row), r = cv.minEnclosingCircle(circle)
            if r > largest_r:
                largest_col, largest_row, largest_r = col, row, r

        return largest_col, largest_row, largest_r

    def snr_by_subtraction(
        self,
        dcm1: pydicom.Dataset,
        dcm2: pydicom.Dataset,
        measured_slice_width=None,
    ):
        """Calculate signal to noise ratio by smoothing

        Args:
            dcm1 (pydicom.Dataset): DICOM image object for signal
            dcm2 (pydicom.Dataset): DICOM image object for noise calculation
            measured_slice_width (float or None): slice width from user input

        Returns:
            tuple of float: SNR and normalised SNR values
        """
        col, row = self.get_object_centre(dcm1)

        difference = np.subtract(
            dcm1.pixel_array.astype("int"), dcm2.pixel_array.astype("int")
        )

        signal = [
            np.mean(roi)
            for roi in self.get_roi_samples(
                ax=None, dcm=dcm1, centre_col=col, centre_row=row
            )
        ]
        noise = np.divide(
            [
                np.std(roi, ddof=1)
                for roi in self.get_roi_samples(
                    ax=None, dcm=difference, centre_col=col, centre_row=row
                )
            ],
            np.sqrt(2),
        )
        snr = np.mean(np.divide(signal, noise))

        normalised_snr = snr * self.get_normalised_snr_factor(
            dcm1, measured_slice_width
        )

        if self.report:
            import matplotlib.pyplot as plt

            fig, axes = plt.subplots(1, 1)
            fig.set_size_inches(5, 5)
            fig.tight_layout(pad=1)

            axes.set_title("difference image")
            axes.imshow(difference, cmap="gray", label="difference image")
            axes.scatter(col, row, 10, marker="+", label="centre")
            self.get_roi_samples(axes, dcm1, col, row)
            axes.legend()

            img_path = os.path.realpath(
                os.path.join(
                    self.report_path,
                    f"{self.img_desc(dcm1)}_snr_subtraction.png",
                )
            )
            fig.savefig(img_path)
            self.report_files.append(img_path)

        return snr, normalised_snr



"""ACR Slice Thickness.

___________________

Reference
_________

`ACR Large Phantom Guidance PDF <https://accreditationsupport.acr.org/helpdesk/attachments/11093487417>`_

Intro
_____

The slice thickness accuracy test assesses the accuracy with which a
 slice of specified thickness is achieved.  The prescribed slice
 thickness is compared with the measured slice thickness.

The ramps appear in a structure called the slice thickness
insert. Figure 10 shows an image of slice 1 with the slice thickness
insert and signal ramps identified. The two ramps are crossed: one has
a negative slope and the other a positive slope with respect to the
plane of slice 1. They are produced by cutting 1 mm wide slots in a
block of plastic. The slots are open to the interior of the phantom
and are filled with the same solution that fills the bulk of the
phantom.

The signal ramps have a slope of 10 to 1 with respect to the plane of
slice 1, which is an angle of about 5.71 deg.  Therefore, the signal
ramps will appear in the image of slice 1 with a length that is 10
times the thickness of the slice. If the phantom is misaligned from
right-left, one ramp will appear longer than the other. The crossed
ramps allow for correction of the error introduced by right-left
misalignment, and the slice thickness formula takes that into account.

ACR Guidelines
______________

ACR Algorithm
+++++++++++++

    #. Display slice 1, and magnify the image by a factor of 2 to 4,
        keeping the slice thickness insert fully visible on the
        screen.
    #. Adjust the display level so that the signal ramps are well
        visualized.
        *. The ramp signal is much lower than that of surrounding
            water, so usually it will be necessary to lower the
            display level substantially and narrow the window.
    #. Place a rectangular ROI at the middle of each ramp as shown
        below in Figure 11.
        *. Note the mean signal values for each of these two ROIs and
         then average those values.
        *. The result is a number approximating the mean signal in the
         middle of the ramps.
        *. An elliptical ROI may be used if a rectangular one is
         unavailable.
    #. Lower the display level to half of the average ramp signal
        calculated in step 3.
        *. Leave the display window set to its minimum.
    #. Use the on-screen distance measurement tool to measure the
        lengths of the top and bottom ramps.  This is illustrated
        below in Figure 12. Record these lengths and compare to the
        action limits.

Computer Vision Methodology
+++++++++++++++++++++++++++

This automated implementation locates the slice thickness insert using
computer vision and calculates ramp widths via curve fitting:

    #. Pre-process image with CLAHE contrast enhancement and Otsu
        binarization to isolate the insert.
    #. Detect contours and identify the slice thickness insert using
       aspect ratio heuristics (the insert is the most elongated
       rectangular structure).
    #. Calculate the oriented bounding box corners of the insert to
        determine ramp geometry.
    #. Construct profile lines across the ramps using geometric
        sublines (spanning 95% of ramp length).
    #. Extract pixel intensity profiles along each line (averaging
        over 4-pixel width).
    #. Smooth profiles with median/Gaussian filters and fit piecewise
        sigmoid functions:
       - Detect the signal peak to separate left and right segments
       - Fit separate sigmoids to each segment using least-squares optimization
       - Blend the sigmoids with a smooth transition weighting
         function at the peak
    #. Calculate FWHM from fitted curves using linear interpolation at
        half-maximum points.
    #. Apply ACR formula: thickness = 0.2 * (L1 * L2) / (L1 + L2),
        accounting for phantom rotation.

Sagittal images are automatically detected and rotated 90 degrees
clockwise prior to analysis.

ACR Scoring Rubric
++++++++++++++++++

The slice thickness is calculated using the following formula:

    Slice thickness = 0.2 x (top x bottom)/(top + bottom)

In the formula above, the `top` and `bottom` are the measured lengths
of the top and bottom signal ramps.

**Note:** 0.2 is a unitless factor that corrects for rotation of the
  phantom about the vertical (y) axis.

For example, if the top signal ramp were 59.5mm long and the bottom
ramp were 47.2mm long, then the calculated slice thickness would be:

    Slice thickness = 0.2 x (59.5mm x 47.2mm)/(59.5mm + 47.2mm) = 5.26 mm.


Notes
-----
..note::

    A failure of this test means that the scanner is producing slices
    of substantially different thickness from the prescribed
    thickness. This problem will generally not occur in isolation
    since the scanner deficiencies that can cause it may also cause
    other image problems. Therefore, the implications of a failure are
    not just that the slices are too thick or thin, but can also
    result in poor image contrast and low SNR.

..warning::

    The automated line placement relies on the insert contour being
    clearly distinguishable from background.  Poor image contrast or
    artifacts may cause the aspect ratio heuristic to fail. The
    fitting algorithm assumes the ramps produce distinct signal peaks;
    very thin slices (<2mm) may challenge the piecewise sigmoid fit.

Documented by Luis M. Santos, M.D.
luis.santos2@nih.gov

and

Alex Drysdale
alexander.drysdale@wales.nhs.uk

"""

# Python Imports
import os
import sys
import traceback
from pathlib import Path

# Module Imports
import cv2
import numpy as np
import pydicom
from scipy.ndimage import gaussian_filter1d
from scipy.optimize import curve_fit
from scipy.signal import find_peaks, medfilt

# Local Imports
from hazenlib.ACRObject import ACRObject
from hazenlib.HazenTask import HazenTask
from hazenlib.logger import logger
from hazenlib.types import Measurement, Result
from hazenlib.utils import XY, Line, Point, get_image_orientation


class ACRSliceThickness(HazenTask):
    """Slice width measurement class for DICOM images of the ACR phantom.

    Automatically detects the slice thickness insert using computer vision,
    places measurement lines across the crossed ramps, and calculates slice
    thickness using piecewise sigmoid fitting to determine FWHM. Handles
    arbitrary phantom rotation and sagittal acquisitions.
    """

    class SignalLine(Line):
        """Subclass of Line to related to the ACR phantom's ramps.

        Extends the Line class with signal profiling and curve fitting
        capabilities specifically designed for measuring the FWHM of
        the characteristic ramp signal (low signal slots in high
        signal background).
        """

        def get_fwhm(self) -> None:
            """Calculate the FWHM by fitting a piecewise sigmoid.

            Fits a smooth piecewise sigmoid curve to the signal
            profile, then calculates the Full Width at Half Maximum
            (FWHM) using linear interpolation at the half-maximum
            crossing points. The fitted curve and FWHM value are
            stored as instance attributes.

            Raises:
                ValueError: If signal has not been computed via
                        get_signal() before calling.

            Sets Attributes:
                fwhm (float): Full width at half maximum in pixels.
                fitted (XY): Fitted curve as XY coordinates.

            """
            if not hasattr(self, "signal"):
                msg = "Signal across line has not been computed!"
                raise ValueError(msg)

            fitted = self._fit_piecewise_sigmoid()

            _, props = find_peaks(
                fitted.y,
                height=0,
                prominence=np.max(fitted.y).item() / 4,
            )
            peak_height = np.max(props["peak_heights"])

            background_l = fitted.y[0]
            background_r = fitted.y[-1]

            half_max_l = (peak_height - background_l) / 2 + background_l
            half_max_r = (peak_height - background_r) / 2 + background_r

            def simple_interpolate(target_y: float, signal: XY) -> float:
                crossing_index = np.where(signal.y > target_y)[0][0]
                x1, x2 = signal.x[crossing_index - 1], signal.x[crossing_index]
                y1, y2 = signal.y[crossing_index - 1], signal.y[crossing_index]
                return x1 + (target_y - y1) * (x2 - x1) / (y2 - y1)

            x_l = simple_interpolate(half_max_l, fitted)
            x_r = simple_interpolate(half_max_r, fitted[:, ::-1])

            self.fwhm = x_r - x_l
            self.fitted = fitted

        def _fit_piecewise_sigmoid(self) -> XY:
            smoothed = self.signal.copy()
            k = round(len(smoothed.y) / 20)
            if k % 2 == 0:
                k += 1
            smoothed.y = medfilt(smoothed.y, k)
            smoothed.y = gaussian_filter1d(smoothed.y, round(k / 2.5))

            peaks, props = find_peaks(
                smoothed.y,
                height=0,
                prominence=np.max(smoothed.y).item() / 4,
            )
            heights = props["peak_heights"]
            peak = peaks[np.argmax(heights)]

            def get_specific_sigmoid(
                whole_data: XY,
                fit_start: int,
                fit_end: int,
            ) -> XY:
                fit_data = whole_data[:, fit_start:fit_end]
                a = np.max(fit_data.y) - np.min(fit_data.y)
                b = np.min(fit_data.y)

                dy = np.diff(fit_data.y)
                dx = np.diff(fit_data.x)
                dx[dx == 0] = 1e-6
                abs_deriv = np.abs(dy / dx)
                abs_grad_max = np.max(abs_deriv)
                k = (
                    np.sign(fit_data.y[-1] - fit_data.y[0])
                    * abs_grad_max
                    * 0.5
                )
                x0 = fit_data.x[np.argmax(abs_deriv)]

                p0 = [a, k, x0, b]

                def sigmoid(
                    x: float | np.ndarray,
                    a: float,
                    k: float,
                    x0: float,
                    b: float,
                ) -> float:
                    exp_term = np.exp(-k * (x - x0))
                    return a / (1 + exp_term) + b

                popt, _ = curve_fit(sigmoid, fit_data.x, fit_data.y, p0=p0)

                def specific_sigmoid(x: float | np.ndarray) -> float:
                    return sigmoid(x, *popt)

                return specific_sigmoid

            sigmoid_l_func = get_specific_sigmoid(smoothed, 0, peak)
            sigmoid_r_func = get_specific_sigmoid(
                smoothed,
                peak,
                len(smoothed.x),
            )

            sigmoid_l = XY(smoothed.x, sigmoid_l_func(smoothed.x))
            sigmoid_r = XY(smoothed.x, sigmoid_r_func(smoothed.x))

            def blending_weight(
                x: np.ndarray,
                transition_x: np.ndarray,
                transition_width: np.ndarray,
            ) -> np.ndarray:
                return 1 / (1 + np.exp(-(x - transition_x) / transition_width))

            w = blending_weight(
                smoothed.x,
                peak,
                1 / 20 * peak + 1 / 20 * (len(smoothed.x) - peak),
            )
            return XY(smoothed.x, (1 - w) * sigmoid_l.y + w * sigmoid_r.y)

    def __init__(self, **kwargs) -> None:
        """Initialise the ACRSliceThickness object."""
        if kwargs.pop("verbose", None) is not None:
            logger.warning(
                "verbose is not a supported argument for %s",
                type(self).__name__,
            )
        super().__init__(**kwargs)
        # Initialise ACR object
        self.ACR_obj = ACRObject(self.dcm_list)

    def run(self) -> Result:
        """Perform slice width measurement.

        Using slice 1 from the ACR phantom image set.

        Returns:
            dict: results are returned in a standardised dictionary
                structure specifying the task name, input DICOM Series
                Description + SeriesNumber + InstanceNumber, task
                measurement key-value pairs, optionally path to the
                generated images for visualisation.

        """
        # Identify relevant slice
        slice_thickness_dcm = self.ACR_obj.slice_stack[0]
        # TODO: image may be 90 degrees cw or acw, could use code to
        # identify which or could be added as extra arg

        ori = get_image_orientation(slice_thickness_dcm)
        if ori == "Sagittal":
            # Get the pixel array from the DICOM file
            img = slice_thickness_dcm.pixel_array

            # Rotate the image 90 degrees clockwise

            rotated_img = np.rot90(img, k=-1)  # k=-1 for 90 degrees clockwise

            # Update the pixel array in the DICOM object
            slice_thickness_dcm.PixelData = rotated_img.tobytes()

        # Initialise results dictionary
        results = self.init_result_dict(desc=self.ACR_obj.acquisition_type())
        results.files = self.img_desc(slice_thickness_dcm)

        try:
            thickness_results = self.get_slice_thickness(slice_thickness_dcm)

            results.add_measurement(
                Measurement(
                    name="SliceWidth",
                    type="measured",
                    subtype="slice width",
                    unit="mm",
                    value=round(thickness_results["thickness"], 2),
                ),
            )

            for ramp in thickness_results["ramps"]:
                results.add_measurement(
                    Measurement(
                        name="SliceWidth",
                        type="measured",
                        subtype="Ramps",
                        unit="mm",
                        value=thickness_results["ramps"][ramp]["width"],
                        visibility="intermediate",
                    ),
                )

        except Exception as e:
            logger.exception(
                "Could not calculate the slice thickness for %s"
                " because of : %s",
                self.img_desc(slice_thickness_dcm),
                e,
            )
            traceback.print_exc(file=sys.stdout)

        # only return reports if requested
        if self.report:
            results.add_report_image(self.report_files)

        return results

    def get_slice_thickness(self, dcm: pydicom.Dataset) -> dict:
        """Measure slice thickness.

        Identify the ramps, measure the line profile, measure the
        FWHM, and use this to calculate the slice thickness.

        Returns:
            dict: Dictionary containing:
                - ``"thickness"`` (float): Calculated slice thickness in mm.
                - ``"ramps"`` (list[float]): FWHM values (in pixels)
                  for each ramp line used in the calculation.

        """
        img = dcm.pixel_array
        lines = self.place_lines(img)
        for line in lines:
            line.get_fwhm()
        slice_thickness = (
            0.2
            * (lines[0].fwhm * lines[1].fwhm)
            / (lines[0].fwhm + lines[1].fwhm)
        )

        if self.report:
            import matplotlib.pyplot as plt  # noqa: PLC0415 I001

            fig, axes = plt.subplots(1, 3, figsize=(16, 8))
            axes[0].imshow(img)
            for i, line in enumerate(lines):
                axes[0].plot(
                    [line.start.x, line.end.x],
                    [line.start.y, line.end.y],
                )
                axes[i + 1].plot(
                    line.signal.x,
                    line.signal.y,
                    label="Raw signal",
                    alpha=0.25,
                    color=f"C{i}",
                )
                axes[i + 1].plot(
                    line.fitted.x,
                    line.fitted.y,
                    label="Fitted piecewise sigmoid",
                    color=f"C{i}",
                )
                axes[i + 1].legend(loc="lower right", bbox_to_anchor=(1, -0.2))

            axes[0].axis("off")

            axes[0].set_title("Plot showing placement of profile lines.")
            axes[1].set_title("Pixel profile across blue line.")
            axes[1].set_xlabel("Distance along blue line (pixels)")
            axes[1].set_ylabel("Pixel value")

            axes[2].set_title("Pixel profile across orange line.")
            axes[2].set_xlabel("Distance along orange line (pixels)")
            axes[2].set_ylabel("Pixel value")
            plt.tight_layout()

            img_path = os.path.realpath(
                Path(self.report_path)
                / f"{self.img_desc(dcm)}_slice_thickness.png",
            )

            fig.savefig(img_path, dpi=600)
            self.report_files.append(img_path)

        return {
            "thickness": slice_thickness,
            "ramps": {
                pos: {"width": line.fwhm}
                for pos, line in zip(("top", "bottom"), lines, strict=True)
            },
        }

    def place_lines(self, img: np.ndarray) -> list["Line"]:
        """Places line on image within ramps insert.

        Works for a rotated phantom.

        Args:
            img (np.ndarray): Pixel array from DICOM image.

        Returns:
            final_lines (list): A list of the two lines as Line objects.

        """
        # Normalize to uint8, enhance contast and binarize using otsu thresh

        img_uint8 = cv2.normalize(img, None, 0, 255, cv2.NORM_MINMAX).astype(
            np.uint8,
        )
        contrast_enhanced = cv2.createCLAHE(
            clipLimit=2.0,
            tileGridSize=(3, 3),
        ).apply(img_uint8)
        _, img_binary = cv2.threshold(
            contrast_enhanced,
            0,
            255,
            cv2.THRESH_BINARY + cv2.THRESH_OTSU,
        )

        # Find contour by x-span sort
        contours, _ = cv2.findContours(
            img_binary.astype(np.uint8),
            mode=cv2.RETR_TREE,
            method=cv2.CHAIN_APPROX_NONE,
        )

        def get_aspect_ratio(contour: np.ndarray) -> float:
            _, (width, height), _ = cv2.minAreaRect(contour)
            return min(width, height) / max(width, height)

        # filter out tiny contours from noise
        thresh_area = 15 * 15
        contours = [
            cont for cont in contours if cv2.contourArea(cont) >= thresh_area
        ]
        # select most elongated insert contour (heuristic for central insert)
        contours_sorted = sorted(
            contours,
            key=lambda c: get_aspect_ratio(c),
        )
        insert_contour = contours_sorted[0]

        # Create list of Point objects for the four corners of the contour
        insert_corners = cv2.boxPoints(cv2.minAreaRect(insert_contour))
        corners = [Point(*p) for p in insert_corners]

        # Define short sides of contours by list of line objects
        corners = sorted(
            corners,
            key=lambda point: corners[0].get_distance_to(point),
        )
        short_sides = [Line(*corners[:2]), Line(*corners[2:])]

        # Get sublines of short sides and force start point to be higher in y
        sublines = [line.get_subline(perc=30) for line in short_sides]
        for line in sublines:
            if line.start.y < line.end.y:
                line.point_swap()

        # Define connecting lines
        connecting_lines = [
            self.SignalLine(sublines[0].start, sublines[1].start),
            self.SignalLine(sublines[0].end, sublines[1].end),
        ]

        # Final lines are sublines of connecting lines
        final_lines = [line.get_subline(perc=95) for line in connecting_lines]
        for line in final_lines:
            line.get_signal(img)

        return final_lines



"""
ACR Sagittal Geometric Accuracy
__________________________________________

Reference
_________

`ACR Large Phantom Guidance PDF <https://accreditationsupport.acr.org/helpdesk/attachments/11093487417>`_

Intro
_____

The ACR Geometric Accuracy Task has an additional task in which you obtain the Sagittal Localizer and measure the
vertical distance from top to bottom of the rectangular shape. The measurement is performed about 15 mm offset from
center per their example. I think the idea is that you can avoid any phantom positional errors that might present
towards the edges of the rectangle. At any rate, this task is a modified version of the axial task.

ACR Guidelines
______________

ACR Algorithm
+++++++++++++

    #. Display the ACR sagittal localizer image. Adjust the display window and level as described below.
    #. Measure the superior to inferior (head to foot) length of the phantom along a line close to the middle
        of the phantom as shown in Figure 2.
                                    ...
    #. Display slice 1 of the ACR T1 series. Adjust the display window and level as described below.
    #. Measure the diameter of the phantom in 2 directions: top-to-bottom and left-to-right (Figure 2).
    #. Determine the window level setting to measure slice 5 of the ACR T1 series, as described below.
        Display slice 5 with that window and level.
    #. Measure the diameter of the phantom in 4 directions: top-to-bottom, left-to-right, and both diagonals.

ACR Scoring Rubric
++++++++++++++++++

Phantom     Sagittal length     Pass/Fail Limit     Axial diameter      Pass/Fail Limit
            (mm)                (mm)                (mm)                (mm)
_______     _______________     _______________     ______________      _______________
Large       148                 148 +/- 3.0         190                 190 +/- 3.0
Medium      134                 134 +/- 2.0         165                 165 +/- 2.0

Notes
_____

.. note::

    Some MR vendors provide the ability to select gradient distortion correction at the operator console. For
    these systems be sure that the distortion correction option is turned on; leaving distortion correction off can
    cause geometric accuracy failure.

.. note::

    A common cause of failure of this test is miscalibration of one or more gradients. A mis-calibrated gradient
    causes its associated image dimension (x, y, or z) to appear longer or shorter than it really is. Mis-calibrated
    gradients also can cause slice position errors. It is normal for gradient calibration to drift over time and to
    require recalibration by the service engineer.

.. note::

    Another possible cause of failure is use of an excessively low acquisition bandwidth. It is common practice
    on low B0 field scanners and at some facilities to reduce acquisition bandwidth, especially on long TE
    acquisitions, to increase signal-to-noise ratio (SNR). This can be pushed to the point that the normal
    inhomogeneities in B0 manifest themselves as spatial distortions in the image. On most scanners the default
    bandwidth for T1-weighted acquisitions is set high enough to avoid this problem. If the geometric accuracy
    test measurements fail, and the ACR T1 series was acquired at low bandwidth, try acquiring that series again
    at a higher bandwidth to see if the problem is eliminated.

Created by Luis M. Santos, M.D.
luis.santos2@nih.gov

2/20/2025
"""

# Python imports
import os
import sys
import traceback

# Module imports
import numpy as np
from hazenlib import logger
from hazenlib.ACRObject import ACRObject
from hazenlib.HazenTask import HazenTask
from hazenlib.types import Measurement


class ACRSagittalGeometricAccuracy(HazenTask):
    """Geometric accuracy measurement class for DICOM images of the ACR phantom."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.ACR_obj = ACRObject(self.dcm_list)

    def run(self) -> dict:
        """Main function for performing geometric accuracy measurement using the first and fifth slices from the
        ACR phantom image set.

        Returns:
            dict: results are returned in a standardised dictionary structure specifying the task name,
                input DICOM Series Description + SeriesNumber + InstanceNumber, task measurement key-value pairs,
                optionally path to the generated images for visualisation.
        """
        dcm = self.ACR_obj.slice_stack[0]
        img_desc = self.img_desc(dcm)

        # Initialise results dictionary
        results = self.init_result_dict(desc=img_desc)

        try:
            lengths_1 = self.get_geometric_accuracy(dcm)
            logger.info(lengths_1)
            results.add_measurement(
                Measurement(
                    name="GeometricAccuracy",
                    value=round(lengths_1, 2),
                    subtype="Sagittal Localizer",
                    unit="mm",
                    description="Superior to Inferior",
                ),
            )
        except Exception as e:
            logger.error(
                f"Could not calculate the geometric accuracy for {self.img_desc(self.ACR_obj.slice_stack[0])} because of : {e}"
            )
            traceback.print_exc(file=sys.stdout)

        # only return reports if requested
        if self.report:
            results.add_report_image(self.report_files)

        return results

    def write_report(self, img, dcm, length_dict, mask, cxy, offset):
        import matplotlib.pyplot as plt

        fig, axes = plt.subplots(3, 1)
        fig.set_size_inches(8, 24)
        fig.tight_layout(pad=4)

        axes[0].imshow(img)
        axes[0].scatter(cxy[0], cxy[1], c="red")
        axes[0].set_title("Centroid Location")

        axes[1].imshow(mask)
        axes[1].set_title("Thresholding Result")

        axes[2].imshow(img)
        axes[2].arrow(
            cxy[0] + offset[0],
            length_dict["Vertical Extent"][0],
            1 + offset[1],
            length_dict["Vertical Extent"][-1]
            - length_dict["Vertical Extent"][0],
            color="orange",
            length_includes_head=True,
            head_width=5,
        )
        axes[2].legend(
            [
                str(np.round(length_dict["Vertical Distance"], 2)) + "mm",
            ]
        )
        axes[2].axis("off")
        axes[2].set_title("Geometric Accuracy for Slice 1")

        img_path = os.path.realpath(
            os.path.join(self.report_path, f"{self.img_desc(dcm)}.png")
        )
        fig.savefig(img_path)
        self.report_files.append(img_path)

    def get_geometric_accuracy(self, dcm):
        """Measures Vertical distance of the Sagittal Localizer.

        Steps
        _____

            #. Grabs slice 1
            #. Creates a mask over the phantom from the pixel array of the DICOM image.
            #. Finds center of phantom
            #. Computes offset from center.
            #. Samples a vertical line profile at that offset.
            #. Computes distance.

        Args:
            dcm (pydicom.Dataset): slice 1

        Returns:
            float: vertical distance.

        """
        img, _, _ = self.ACR_obj.get_presentation_pixels(dcm)

        cxy, _ = self.ACR_obj.find_phantom_center(
            img,
            self.ACR_obj.dx,
            self.ACR_obj.dy,
            axial=False,
        )
        offset = (int(np.round(-15 / self.ACR_obj.dx)), 0)

        length_dicts = []
        for threshold in np.linspace(0.05, 0.09, 10, endpoint=True):
            mask = self.ACR_obj.get_mask_image(
                img,
                cxy,
                mag_threshold=threshold,
            )
            length_dicts.append(
                self.ACR_obj.measure_orthogonal_lengths(
                    mask,
                    cxy,
                    v_offset=offset,
                ),
            )

        length_dict = {}
        for key in ("Vertical Distance",):
            length_dict[key] = np.mean([d[key] for d in length_dicts])

        if self.report:
            length_dict["Vertical Extent"] = length_dicts[
                len(length_dicts) // 2
            ]["Vertical Extent"]
            self.write_report(img, dcm, length_dict, mask, cxy, offset)

        return length_dict["Vertical Distance"]



"""
Uniformity + Ghosting & Distortion

Calculates uniformity for a single-slice image of a uniform MRI phantom

This script implements the IPEM/MAGNET method of measuring fractional uniformity.
It also calculates integral uniformity using a 75% area FOV ROI and CoV for the same ROI.

This script also measures Ghosting within a single image of a uniform phantom.
This follows the guidance from ACR for testing their large phantom.

A simple measurement of distortion is also made by comparing the height and width of the circular phantom.

Created by Neil Heraghty
neil.heraghty@nhs.net

14/05/2018

"""

import os
import sys
import traceback

import hazenlib.exceptions as exc
import hazenlib.utils
import numpy as np
from hazenlib.HazenTask import HazenTask
from hazenlib.logger import logger
from hazenlib.types import Measurement, Result


class Uniformity(HazenTask):
    """Uniformity measurement class for DICOM images of the MagNet phantom

    Inherits from HazenTask class
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # Set the single DICOM input to be the first in the list
        self.single_dcm = self.dcm_list[0]

    def run(self) -> Result:
        """Main function for performing uniformity measurement

        Returns:
            dict: results are returned in a standardised dictionary structure specifying the task name, input DICOM Series Description + SeriesNumber + InstanceNumber, task measurement key-value pairs, optionally path to the generated images for visualisation
        """
        results = self.init_result_dict()
        results.files = self.img_desc(self.single_dcm)

        try:
            horizontal_uniformity, vertical_uniformity = (
                self.get_fractional_uniformity(self.single_dcm)
            )
            results.add_measurement(
                Measurement(
                    name="Uniformity",
                    type="measured",
                    subtype="Horizontal",
                    unit="%",
                    value=round(horizontal_uniformity, 2),
                ),
            )
            results.add_measurement(
                Measurement(
                    name="Uniformity",
                    type="measured",
                    subtype="Vertical",
                    unit="%",
                    value=round(vertical_uniformity, 2),
                ),
            )
        except Exception as e:
            logger.exception(
                "Could not calculate the uniformity for %s because of : %s",
                self.img_desc(self.single_dcm),
                e,
            )
            traceback.print_exc(file=sys.stdout)

        # only return reports if requested
        if self.report:
            results.add_report_image(self.report_files)

        return results

    def mode(self, a, axis=0):
        """Finds the modal value of an array. From scipy.stats.mode

        Args:
            a (np.array): _description_
            axis (int, optional): Axis to calculate mode along. Defaults to 0.

        Returns:
            most_frequent: the modal value
            old_counts: the number of times this value was counted (check this)
        """
        scores = np.unique(np.ravel(a))  # get ALL unique values
        test_shape = list(a.shape)
        test_shape[axis] = 1
        old_most_frequent = np.zeros(test_shape)
        old_counts = np.zeros(test_shape)
        most_frequent = None

        for score in scores:
            template = a == score
            counts = np.expand_dims(np.sum(template, axis), axis)
            most_frequent = np.where(
                counts > old_counts, score, old_most_frequent
            )
            old_counts = np.maximum(counts, old_counts)
            old_most_frequent = most_frequent

        return most_frequent, old_counts

    def get_object_centre(self, dcm):
        """Locate centre coordinates

        Args:
            dcm (pydicom.FileDataset): DICOM image object

        Raises:
            Exception: _description_

        Returns:
            tuple: x and y coordinates
        """
        arr = dcm.pixel_array
        shape_detector = hazenlib.utils.ShapeDetector(arr=arr)
        orientation = hazenlib.utils.get_image_orientation(dcm)

        if orientation in ["Sagittal", "Coronal"]:
            # orientation is sagittal to patient
            try:
                (x, y), size, angle = shape_detector.get_shape("rectangle")
            except exc.ShapeError:
                raise

        elif orientation == "Transverse":
            # orientation is axial
            x, y, r = shape_detector.get_shape("circle")

        else:
            raise Exception(
                "Direction must be Transverse, Sagittal or Coronal."
            )

        return int(x), int(y)

    def get_fractional_uniformity(self, dcm):
        """Get fractional uniformity

        Args:
            dcm (pydicom.FileDataset): DICOM image object

        Returns:
            tuple: values of horizontal and vertical fractional uniformity
        """
        arr = dcm.pixel_array
        x, y = self.get_object_centre(dcm)

        central_roi = arr[(y - 5) : (y + 5), (x - 5) : (x + 5)].flatten()
        # Create central 10x10 ROI and measure modal value

        central_roi_mode, mode_popularity = self.mode(central_roi)

        # Create 160-pixel profiles (horizontal and vertical, centred at x,y)
        horizontal_roi = arr[(y - 5) : (y + 5), (x - 80) : (x + 80)]
        horizontal_profile = np.mean(horizontal_roi, axis=0)
        vertical_roi = arr[(y - 80) : (y + 80), (x - 5) : (x + 5)]
        vertical_profile = np.mean(vertical_roi, axis=1)

        # Count how many elements are within 0.9-1.1 times the modal value
        horizontal_count = np.where(
            np.logical_and(
                (horizontal_profile > (0.9 * central_roi_mode)),
                (horizontal_profile < (1.1 * central_roi_mode)),
            )
        )
        horizontal_count = len(horizontal_count[0])
        vertical_count = np.where(
            np.logical_and(
                (vertical_profile > (0.9 * central_roi_mode)),
                (vertical_profile < (1.1 * central_roi_mode)),
            )
        )
        vertical_count = len(vertical_count[0])

        # Calculate fractional uniformity
        fractional_uniformity_horizontal = horizontal_count / 160
        fractional_uniformity_vertical = vertical_count / 160

        if self.report:
            import matplotlib.pyplot as plt
            from matplotlib.collections import PatchCollection
            from matplotlib.patches import Rectangle

            fig, ax = plt.subplots()
            rects = [
                Rectangle(
                    (x - 5, y - 5),
                    10,
                    10,
                    facecolor="None",
                    edgecolor="red",
                    linewidth=3,
                ),
                Rectangle(
                    (x - 80, y - 5),
                    160,
                    10,
                    facecolor="None",
                    edgecolor="green",
                ),
                Rectangle(
                    (x - 5, y - 80),
                    10,
                    160,
                    facecolor="None",
                    edgecolor="yellow",
                ),
            ]
            pc = PatchCollection(rects, match_original=True)
            ax.imshow(arr, cmap="gray")
            ax.add_collection(pc)
            ax.scatter(x, y, 5)

            img_path = os.path.realpath(
                os.path.join(self.report_path, f"{self.img_desc(dcm)}.png")
            )
            fig.savefig(img_path)
            self.report_files.append(img_path)

        return fractional_uniformity_horizontal, fractional_uniformity_vertical



"""
ACR Ghosting

https://www.acraccreditation.org/-/media/acraccreditation/documents/mri/largephantomguidance.pdf

Calculates the percent-signal ghosting for slice 7 of the ACR phantom.

This script calculates the percentage signal ghosting in accordance with the ACR Guidance.
This is done by first defining a large 200cm2 ROI before placing 10cm2 elliptical ROIs outside the phantom along the
cardinal directions. The results are also visualised.

Created by Yassine Azma
yassine.azma@rmh.nhs.uk

14/11/2022
"""

import os
import sys
import traceback

import numpy as np
from hazenlib.ACRObject import ACRObject
from hazenlib.HazenTask import HazenTask
from hazenlib.logger import logger
from hazenlib.types import Measurement, Result


class ACRGhosting(HazenTask):
    """Ghosting measurement class for DICOM images of the ACR phantom."""

    def __init__(self, **kwargs):
        if kwargs.pop("verbose", None) is not None:
            logger.warning(
                "verbose is not a supported argument for %s",
                type(self).__name__,
            )
        super().__init__(**kwargs)
        # Initialise ACR object
        self.ACR_obj = ACRObject(self.dcm_list)

    def run(self) -> Result:
        """Main function for performing ghosting measurement using slice 7 from the ACR phantom image set.

        Returns:
            dict: results are returned in a standardised dictionary structure specifying the task name, input DICOM Series Description + SeriesNumber + InstanceNumber, task measurement key-value pairs, optionally path to the generated images for visualisation.
        """
        # Initialise results dictionary
        results = self.init_result_dict(desc=self.ACR_obj.acquisition_type())
        results.files = self.img_desc(self.ACR_obj.slice_stack[6])

        try:
            result = self.get_signal_ghosting(self.ACR_obj.slice_stack[6])
            results.add_measurement(
                Measurement(
                    name="Ghosting",
                    type="measured",
                    subtype="signal ghosting",
                    value=round(result, 3),
                    unit="%",
                ),
            )

        except Exception as e:
            logger.exception(
                "Could not calculate the percent-signal ghosting for %s"
                "because of : %s",
                self.img_desc(self.ACR_obj.slice_stack[6]),
                e,
            )
            traceback.print_exc(file=sys.stdout)

        # only return reports if requested
        if self.report:
            results.add_report_image(self.report_files)

        return results

    def get_signal_ghosting(self, dcm):
        """Calculate the percentage signal ghosting (PSG). \n
        Sample signal intensity from ellipses outside the phantom in
        four directions and calculate the mean signal value within each.
        Percentage signal ghosting (PSG) is then expressed as the mean signal in these four ROIs
        as a percentage of the mean signal in a ROI in the centre of the phantom.

        Args:
            dcm (pydicom.Dataset): DICOM image object.

        Returns:
            float: percentage ghosting value.
        """
        img = dcm.pixel_array
        # Required pixel radius to produce ~200cm2 ROI
        r_large = np.ceil(80 / self.ACR_obj.dx).astype(int)
        dims = img.shape

        (centre_x, centre_y), _ = self.ACR_obj.find_phantom_center(
            img, self.ACR_obj.dx, self.ACR_obj.dy
        )
        mask = self.ACR_obj.get_mask_image(img, (centre_x, centre_y))

        nx = np.linspace(1, dims[0], dims[0])
        ny = np.linspace(1, dims[1], dims[1])

        x, y = np.meshgrid(nx, ny)

        lroi = np.square(x - centre_x) + np.square(
            y - centre_y - np.divide(5, self.ACR_obj.dy)
        ) <= np.square(r_large)
        # Short axis diameter for an ellipse of 10cm2 with a 1:4 axis ratio
        sad = 2 * np.ceil(np.sqrt(1000 / (4 * np.pi)) / self.ACR_obj.dx)

        # WEST ELLIPSE
        # find first column in mask
        w_point = np.argwhere(np.sum(mask, 0) > 0)[0]
        # initialise centre of ellipse
        w_centre = [centre_y, np.floor(w_point / 2)]
        # edge of ellipse towards left FoV (+ tolerance)
        left_fov_to_centre = w_centre[1] - sad / 2 - 5
        # edge of ellipse towards left side of phantom (+ tolerance)
        centre_to_left_phantom = w_centre[1] + sad / 2 + 5
        if left_fov_to_centre < 0 or centre_to_left_phantom > w_point:
            diffs = [left_fov_to_centre, centre_to_left_phantom - w_point]
            ind = diffs.index(max(diffs, key=abs))
            # ellipse scaling factor
            w_factor = (sad / 2) / (sad / 2 - np.absolute(diffs[ind]))
        else:
            w_factor = 1

        # generate ellipse mask
        w_ellipse = np.square((y - w_centre[0]) / (4 * w_factor)) + np.square(
            (x - w_centre[1]) * w_factor
        ) <= np.square(10 / self.ACR_obj.dx)

        # EAST ELLIPSE
        # find last column in mask
        e_point = np.argwhere(np.sum(mask, 0) > 0)[-1]
        # initialise centre of ellipse
        e_centre = [
            centre_y,
            e_point + np.ceil((dims[1] - e_point) / 2),
        ]
        # edge of ellipse towards right FoV (+ tolerance)
        right_fov_to_centre = e_centre[1] + sad / 2 + 5
        # edge of ellipse towards right side of phantom (+ tolerance)
        centre_to_right_phantom = e_centre[1] - sad / 2 - 5
        if (
            right_fov_to_centre > dims[1] - 1
            or centre_to_right_phantom < e_point
        ):
            diffs = [
                dims[1] - 1 - right_fov_to_centre,
                centre_to_right_phantom - e_point,
            ]
            ind = diffs.index(max(diffs, key=abs))
            # ellipse scaling factor
            e_factor = (sad / 2) / (sad / 2 - np.absolute(diffs[ind]))
        else:
            e_factor = 1

        # generate ellipse mask
        e_ellipse = np.square((y - e_centre[0]) / (4 * e_factor)) + np.square(
            (x - e_centre[1]) * e_factor
        ) <= np.square(10 / self.ACR_obj.dx)

        # NORTH ELLIPSE
        # find first row in mask
        n_point = np.argwhere(np.sum(mask, 1) > 0)[0]
        # initialise centre of ellipse
        n_centre = [np.round(n_point / 2), centre_x]
        # edge of ellipse towards top FoV (+ tolerance)
        top_fov_to_centre = n_centre[0] - sad / 2 - 5
        # edge of ellipse towards top side of phantom (+ tolerance)
        centre_to_top_phantom = n_centre[0] + sad / 2 + 5
        if top_fov_to_centre < 0 or centre_to_top_phantom > n_point:
            diffs = [top_fov_to_centre, centre_to_top_phantom - n_point]
            ind = diffs.index(max(diffs, key=abs))
            # ellipse scaling factor
            n_factor = (sad / 2) / (sad / 2 - np.absolute(diffs[ind]))
        else:
            n_factor = 1

        # generate ellipse mask
        n_ellipse = np.square((y - n_centre[0]) * n_factor) + np.square(
            (x - n_centre[1]) / (4 * n_factor)
        ) <= np.square(10 / self.ACR_obj.dx)
        # SOUTH ELLIPSE
        # find last row in mask
        s_point = np.argwhere(np.sum(mask, 1) > 0)[-1]
        # initialise centre of ellipse
        s_centre = [s_point + np.round((dims[1] - s_point) / 2), centre_x]
        # edge of ellipse towards bottom FoV (+ tolerance)
        bottom_fov_to_centre = s_centre[0] + sad / 2 + 5
        # edge of ellipse towards
        centre_to_bottom_phantom = s_centre[0] - sad / 2 - 5
        if (
            bottom_fov_to_centre > dims[0] - 1
            or centre_to_bottom_phantom < s_point
        ):
            diffs = [
                dims[0] - 1 - bottom_fov_to_centre,
                centre_to_bottom_phantom - s_point,
            ]
            ind = diffs.index(max(diffs, key=abs))
            # ellipse scaling factor
            s_factor = (sad / 2) / (sad / 2 - np.absolute(diffs[ind]))
        else:
            s_factor = 1

        s_ellipse = np.square((y - s_centre[0]) * s_factor) + np.square(
            (x - s_centre[1]) / (4 * s_factor)
        ) <= np.square(10 / self.ACR_obj.dx)

        large_roi_val = np.mean(img[np.nonzero(lroi)])
        w_ellipse_val = np.mean(img[np.nonzero(w_ellipse)])
        e_ellipse_val = np.mean(img[np.nonzero(e_ellipse)])
        n_ellipse_val = np.mean(img[np.nonzero(n_ellipse)])
        s_ellipse_val = np.mean(img[np.nonzero(s_ellipse)])

        psg = 100 * np.absolute(
            ((n_ellipse_val + s_ellipse_val) - (w_ellipse_val + e_ellipse_val))
            / (2 * large_roi_val)
        )

        if self.report:
            import matplotlib.pyplot as plt

            fig, axes = plt.subplots(2, 1)
            fig.set_size_inches(8, 16)
            fig.tight_layout(pad=4)

            theta = np.linspace(0, 2 * np.pi, 360)

            axes[0].imshow(img)
            axes[0].scatter(centre_x, centre_y, c="red")
            axes[0].axis("off")
            axes[0].set_title("Centroid Location")

            axes[1].imshow(img)
            axes[1].plot(
                r_large * np.cos(theta) + centre_x,
                r_large * np.sin(theta) + centre_y + 5 / self.ACR_obj.dy,
                c="black",
            )
            axes[1].text(
                centre_x - 3 * np.floor(10 / self.ACR_obj.dx),
                centre_y + np.floor(10 / self.ACR_obj.dy),
                "Mean = " + str(np.round(large_roi_val, 2)),
                c="white",
            )

            axes[1].plot(
                10.0 / self.ACR_obj.dx * np.cos(theta) / w_factor
                + w_centre[1],
                10.0 / self.ACR_obj.dx * np.sin(theta) * 4 * w_factor
                + w_centre[0],
                c="red",
            )
            axes[1].text(
                w_centre[1] - np.floor(10 / self.ACR_obj.dx),
                w_centre[0],
                "Mean = " + str(np.round(w_ellipse_val, 2)),
                c="white",
            )

            axes[1].plot(
                10.0 / self.ACR_obj.dx * np.cos(theta) / e_factor
                + e_centre[1],
                10.0 / self.ACR_obj.dx * np.sin(theta) * 4 * e_factor
                + e_centre[0],
                c="red",
            )
            axes[1].text(
                e_centre[1] - np.floor(30 / self.ACR_obj.dx),
                e_centre[0],
                "Mean = " + str(np.round(e_ellipse_val, 2)),
                c="white",
            )

            axes[1].plot(
                10.0 / self.ACR_obj.dx * np.cos(theta) * 4 * n_factor
                + n_centre[1],
                10.0 / self.ACR_obj.dx * np.sin(theta) / n_factor
                + n_centre[0],
                c="red",
            )
            axes[1].text(
                n_centre[1] - 5 * np.floor(10 / self.ACR_obj.dx),
                n_centre[0],
                "Mean = " + str(np.round(n_ellipse_val, 2)),
                c="white",
            )

            axes[1].plot(
                10.0 / self.ACR_obj.dx * np.cos(theta) * 4 * s_factor
                + s_centre[1],
                10.0 / self.ACR_obj.dx * np.sin(theta) / s_factor
                + s_centre[0],
                c="red",
            )
            axes[1].text(
                s_centre[1],
                s_centre[0],
                "Mean = " + str(np.round(s_ellipse_val, 2)),
                c="white",
            )

            axes[1].axis("off")
            axes[1].set_title(
                "Percent Signal Ghosting = " + str(np.round(psg, 3)) + "%"
            )
            img_path = os.path.realpath(
                os.path.join(self.report_path, f"{self.img_desc(dcm)}.png")
            )
            fig.savefig(img_path)
            self.report_files.append(img_path)

        return psg



"""
ACR Spatial Resolution (ACR)
____________________________

Reference Materials
+++++++++++++++++++

    *. `Large Phantom Instructions <https://www.acraccreditation.org/-/media/acraccreditation/documents/mri/largephantomguidance.pdf>`_

Intro
+++++

For this test, resolution in slice 1 of each of the 2 ACR axial series is evaluated. The following procedure is
repeated for each of those series:

    #. Display the slice 1 image.
    #. Magnify the image by a factor of between 2 and 4, keeping the resolution insert visible in the display.
        This is illustrated in Figure 8.
    #. Begin with the leftmost pair of hole arrays, which is the pair with the largest hole size, 1.1 mm.
    #. Look at the rows of holes in the UL array, and adjust the display window and level to best show the holes as
        distinct from one another.
    #. If all 4 holes in any single row are distinguishable from one another, score the image as resolved right-to-left
        at this particular hole size.

To be  distinguishable  or resolved, it is not necessary that image intensity drop to zero between the holes.
To be distinguishable a single window and level setting can be found such that **all 4 holes in at least one row**
are recognizable as points of brighter signal intensity than the spaces between them. Figure 9a shows the typical
appearance of well-resolved holes.

..warning::

        2.5 Causes of Failure and Corrective Actions

        Excessive image filtering can cause failure. Many types of filtering that are used to make the images appear less
        noisy also smooth the image, which blurs small structures. A site that has failed the high-contrast resolution test
        should check that any user selectable image filtering is either turned off, or at least set to the low end of the
        available filter settings.

        Poor eddy current compensation can cause failure. The scanner s service engineer should check and adjust the eddy
        current compensation if this problem is suspected

Based on the above excerpt from the ACR Guidance Manual, we just need to find one row in which all 4 dots are high
intensity.

Steps
+++++

    #. Create rectangular ROI.
    #. Window Level (which I implemented for the ACR Low Contrast Object Detectability Task).
    #. Threshold.
    #. Identify center to corner pixel.
    #. Build line profile in this row or column.
    #. Find 4 peaks.
    #. Confirm peaks are spaced out.
    #. Confirm that line is valid.
    #. Stop.
    #. Repeat steps for lower box.


Created by Luis M. Santos, M.D.
luis.santos2@nih.gov

02/06/2025
"""

import os
import sys
import traceback

import numpy as np
from hazenlib.ACRObject import ACRObject
from hazenlib.HazenTask import HazenTask
from hazenlib.logger import logger
from hazenlib.types import Measurement, Result
from hazenlib.utils import wait_on_parallel_results


class ACRSpatialResolution(HazenTask):
    """Spatial resolution measurement class for DICOM images of the ACR phantom

    Inherits from HazenTask class
    """

    ROI_OFFSET = 23  #: 23mm separation between ROIs
    BASE_X_OFFSET = -14  #: -14mm from centroid for 1.1mm resolution array
    BASE_Y_OFFSET = 40  #: 40mm from centroid for 1.1mm resolution array
    DEFAULT_GROUP_SIZE = 2  #: If a dataset is a 1mm resolution, the crop roi is sized such that we can check a row's
    #: value by looking at pairs of rows in data

    def __init__(self, **kwargs):
        if kwargs.pop("verbose", None) is not None:
            logger.warning(
                "verbose is not a supported argument for %s",
                type(self).__name__,
            )
        super().__init__(**kwargs)
        self.ACR_obj = ACRObject(self.dcm_list)

    def run(self) -> Result:
        """Main function for performing spatial resolution measurement
        using slice 1 from the ACR phantom image set

        Returns:
            dict: results are returned in a standardised dictionary structure specifying the task name,
            input DICOM Series Description + SeriesNumber + InstanceNumber, task measurement key-value pairs,
            optionally path to the generated images for visualisation
        """
        # Identify relevant slices
        dcm = self.ACR_obj.slice_stack[0]

        # Initialise results dictionary
        results = self.init_result_dict(desc=self.ACR_obj.acquisition_type())
        results.files = self.img_desc(dcm)

        try:
            detected_rows = self.get_spatially_resolved_rows(dcm)
            hole_arrays = self.get_resolved_arrays(detected_rows)

            results.add_measurement(
                Measurement(
                    name="SpatialResolution",
                    type="measured",
                    subtype="Hole Arrays",
                    unit="mm",
                    value=hole_arrays,
                    visibility="intermediate",
                ),
            )

            for res, desc, visibility in zip(
                self.get_best_resolution(hole_arrays),
                ("Upper Left", "Lower Right", "Best"),
                ("intermediate", "intermediate", "final"),
                strict=True,
            ):
                results.add_measurement(
                    Measurement(
                        name="SpatialResolution",
                        type="measured",
                        subtype=desc,
                        unit="mm",
                        value=res,
                        visibility=visibility,
                    ),
                )

        except Exception as e:
            logger.exception(
                "Could not calculate the spatial resolution for %s"
                " because of : %s",
                self.img_desc(dcm),
                e,
            )
            traceback.print_exc(file=sys.stdout)

        # only return reports if requested
        if self.report:
            results.add_report_image(self.report_files)

        return results

    def write_report(
        self, dcm, img, center, roi_coords, processed_rois, width
    ):
        import matplotlib.patches as patches
        import matplotlib.pyplot as plt

        fig, axes = plt.subplot_mosaic(
            [
                ["main", "main", "main", "main", "main"],
                ["main", "main", "main", "main", "main"],
                ["main", "main", "main", "main", "main"],
                ["main", "main", "main", "main", "main"],
                ["main", "main", "main", "main", "main"],
                [".", ".", ".", ".", "."],
                [".", "ul1", "ul2", "ul3", "."],
                [".", ".", ".", ".", "."],
                [".", ".", ".", ".", "."],
            ],
            layout="constrained",
            per_subplot_kw={"main": {"xbound": (0, 750), "ybound": (0, 500)}},
        )
        fig.set_size_inches(8, 10)
        # fig.tight_layout(pad=1)

        axes["main"].imshow(img, interpolation="none")
        for i in range(len(roi_coords)):
            roi_center = roi_coords[i]
            rect = patches.Rectangle(
                (roi_center[0] - width / 2, roi_center[1] - width / 2),
                width,
                width,
                linewidth=1,
                edgecolor="w",
                facecolor="none",
            )
            axes["main"].add_patch(rect)
        axes["main"].scatter(center[0], center[1], c="red", s=1)
        axes["main"].axis("off")
        axes["main"].set_title("Centroid + ROI Placement")

        roi_center = roi_coords[0]
        axes["ul1"].annotate(
            "UL",
            (-30, roi_center[1] / 2.5),
            xycoords="axes pixels",
            fontsize="large",
        )
        axes["ul1"].annotate(
            "LR",
            (-30, roi_center[1] / 5),
            xycoords="axes pixels",
            fontsize="large",
        )

        axes["ul1"].set_title("1.1mm")
        axes["ul2"].set_title("1.0mm")
        axes["ul3"].set_title("0.9mm")

        for i in range(len(processed_rois)):
            ul = processed_rois[i][1]
            indx = i + 1
            ul_name = f"ul{indx}"
            axes[ul_name].imshow(ul, interpolation="none")
            axes[ul_name].axis("off")
            axes[ul_name].set_xlabel(ul_name.upper())

        img_path = os.path.realpath(
            os.path.join(self.report_path, f"{self.img_desc(dcm)}.png")
        )
        fig.savefig(img_path)
        self.report_files.append(img_path)

    def get_resolved_arrays(self, detected_rows):
        """Iterates through the detected roi scores. Since these come in UL/LR pairs, the presence of -1 in any of the
        paired items disqualifies the hole array from being a resolved image. Go from 1.1 to 0.9 array and record
        which arrays were resolved. Return this struct.

        Args:
            detected_rows (list of int): CList of ints describing each roi's resolved row.

        Returns:
            dict: Hole array results.
        """
        resolved_arrays = {}
        for i in range(0, len(detected_rows), self.DEFAULT_GROUP_SIZE):
            ul = detected_rows[i]
            lr = detected_rows[i + 1]
            resolved = min(ul, lr) != -1
            resolution = np.round(
                1.1 - 0.1 * int(i / self.DEFAULT_GROUP_SIZE), 1
            )
            resolved_arrays.update(
                {resolution: {"UL": ul, "LR": lr, "resolved": resolved}}
            )
        return resolved_arrays

    def get_best_resolution(self, hole_arrays):
        """Iterates through the hole arrays and extract which of the resolved resolutions is best.

        Args:
            hole_arrays (dict): Array resolution struct results.

        Returns:
            tuple of floats:

                #. Highest resolution among Upper Left ROIs.
                #. Highest resolution among Lower Right ROIs.
                #. Highest resolution from a set of 1.1, 1.0, and 0.9.
        """
        ul_resolution = 99
        lr_resolution = 99
        best_resolution = 99
        for k, v in hole_arrays.items():
            ul = v["UL"]
            lr = v["LR"]
            if ul > -1 and k < ul_resolution:
                ul_resolution = k
            if lr > -1 and k < lr_resolution:
                lr_resolution = k
            if v["resolved"]:
                best_resolution = k
        return ul_resolution, lr_resolution, best_resolution

    def get_processed_roi(self, img, width, height, loc):
        """Takes an input image and applies a series of preprocessing steps meant to optimize the output for robust
        detection of signal peaks.

        Preprocessing Steps
        ___________________

            #. Place small ROI using offset from center.
            #. Crop image at ROI.
            #. Apply linear windowing method using ROI pixel population.
            #. Normalize the image to 1 (makes each pixel Boolean compatible).

        Args:
            img (np.ndarray|np.ma.MaskedArray): Cropped image data
            width (int): Width of rectangular ROI.
            height (int): Height of rectangular ROI.
            loc (tuple of int): Center of where the ROI is expected to be placed.

        Returns:
            tuple of np.ndarray: Denoised windowed output and normalized output.
        """
        roi_x, roi_y = loc

        roi = self.ACR_obj.crop_image(img, roi_x, roi_y, width, height)

        # Windowing step
        center, window_width = self.ACR_obj.compute_center_and_width(roi)
        logger.info(f"Applying center {center} and width 0!")
        leveled = self.ACR_obj.apply_window_center_width(roi, center, 0)

        # Return denoised and normalized (0 to 1) results.
        return leveled, self.ACR_obj.normalize(leveled, 1)

    def find_resolved_row(self, roi, grouping_size=2, ul=True):
        """Goes row by row and detects the number of intensity peaks present.
        To make accuracy robust to signal that bleeds into an otherwise empty row, I check for the max count in pairs of
        rows starting with the first row that contains any signal.

        Detection Steps
        _______________

            #. Rotate ROI 90deg if LR, otherwise use ROI as is.
            #. Iterate row by row and collect the count of signal peaks present.
            #. Trim leading zeros.
            #. From top of signal, iterate pair by pair of rows.
            #. Select the max count of peaks at each row pair.
            #. Trim trailing zeros.
            #. Iterate through each count and stop at the first place with a count of 4.
            #. Report this index location as index + 1. That is the human expected row where the input was deemed resolved.
            #. Otherwise, report -1 to indicate we did not find any resolved rows!


        Args:
            roi (np.ndarray|np.ma.MaskedArray): Cropped image data
            grouping_size (int): How many data rows represent a single "true" result row. Defaults to 2 for a 1mm.
            ul (bool): Whether the input is the UL or LR ROI. We want to rotate the LR input such that we can apply the
                        same row traversal

        Returns:
            int: Row number expressing which row was detected as containing 4 peaks. For UL, this number reflect the
                row number starting at the top. For LR, this number reflects the column number starting at the right
                most column and moving to the left (in). Defaults to -1 to demarcate that no rows met the detection
                criteria.
        """
        roi = roi if ul else np.rot90(roi)
        row_length = roi.shape[1]
        vals = np.zeros((roi.shape[0], 1))
        for i in range(len(vals)):
            peaks, intensities = self.ACR_obj.find_n_highest_peaks(
                roi[i], row_length
            )
            vals[i] = len(peaks)
        vals = np.trim_zeros(vals, "f")
        vals = [
            np.max(vals[i : i + grouping_size]).astype(np.int_)
            for i in range(0, len(vals), grouping_size)
        ]
        vals = np.trim_zeros(vals, "b")
        vals = vals[
            0:3
        ]  # We should only have 4 rows; anything else is an artefact and cannot be relied on as true.
        # If we have not resolved any rows within the first 4, this array is probably unresolvable.
        # Ignore the 4th row because we now keep the full array (ul + lr components) to avoid
        # interpolation artefacts. As a result, the 4th row contains spots for both the ul and lr
        # components which can make an array pass inadvertently. ALso, has the added benefit of
        # enforcing strictness here which can only be a good thing in terms of exceeding quality
        # controls prescribed by the ACR.
        logger.info(f"Row spots detected {vals}")
        for i in range(len(vals)):
            if vals[i] == 4:
                return i + 1
        return -1

    def get_spatially_resolved_rows(self, dcm):
        """Generates a series of ROIs centered around the hole arrays present in the ACR phantom.
        Preprocesses these rois and then attempts to detect hole array.
        These steps are parallelized for maximum processing efficiency.
        Afterward, write a report if requested. The report has the big picture view on the roi placements and includes
        the preprocessing outputs.

        Args:
            dcm (pydicom.Dataset): DICOM image object

        Returns:
            list of int: Array containing 6 values. One value for each ROI. The format is [ul, lr, ul, lr, ul, lr].
                        Each pair of rois denotes one of three resolution arrays ([1.1, 1.0, 0.9]). Each value
                        represents the row or column in which we detected 4 distinguishable holes.
        """
        img, rescaled, presentation = self.ACR_obj.get_presentation_pixels(dcm)

        # Detect Phantom center.
        dx, dy = float(self.ACR_obj.dx), float(self.ACR_obj.dy)
        cxy, _ = self.ACR_obj.find_phantom_center(img, dx, dy)
        width = int(np.round(25 / dx))

        logger.info(f"Center           => {cxy}")
        logger.info(f"Pixel Resolution => {dx, dy}")
        logger.info(f"ROI width        => {width}")

        # Generate preprocessed ROIs
        task_args = []
        roi_coords = []
        for i in range(3):
            # Request UL ROI
            x_off = (self.BASE_X_OFFSET / dx) + i * self.ROI_OFFSET / dx
            x, y = (int(cxy[0] + x_off), int(cxy[1] + self.BASE_Y_OFFSET / dy))
            logger.info(f"ROI Center for Array {i} => {x, y}")
            task_args.append((rescaled, width, width, (x, y)))
            roi_coords.append((x, y))

        processed_rois = wait_on_parallel_results(
            self.get_processed_roi, task_args
        )

        # Detect resolved row/column in each ROI.
        group_size = int(np.round(self.DEFAULT_GROUP_SIZE / dx))
        task_args.clear()
        for i in range(len(processed_rois)):
            roi = processed_rois[i][1]
            task_args.append((roi, group_size, True))
            task_args.append((roi, group_size, False))
        detected_rows = wait_on_parallel_results(
            self.find_resolved_row, task_args
        )

        logger.info(f"Detected rows => {detected_rows}")

        if self.report:
            self.write_report(dcm, img, cxy, roi_coords, processed_rois, width)

        return detected_rows



"""
ACR Slice Position

https://www.acraccreditation.org/-/media/acraccreditation/documents/mri/largephantomguidance.pdf

Calculates the bar length difference for slices 1 and 11 of the ACR phantom.

This script calculates the bar length difference in accordance with the ACR Guidance. Line profiles are drawn
vertically through the left and right wedges. The right wedge's line profile is shifted and wrapped round before being
subtracted from the left wedge's line profile, e.g.:

Right line profile: [1, 2, 3, 4, 5] \n
Right line profile wrapped round by 1: [2, 3, 4, 5, 1]

This wrapping process, from hereon referred to as 'circular shifting', is then used for subtractions.

The shift used to produce the minimum difference between the circularly shifted right line profile and the static left
one is used to determine the bar length difference, which is twice the slice position displacement. \n
The results are also visualised.

Created by Yassine Azma
yassine.azma@rmh.nhs.uk

28/12/2022

Refactored by Luis M. Santos
luis.santos2@nih.gov

07/02/2025
"""

import os
import sys
import traceback

import numpy as np
import scipy
import skimage.measure
import skimage.morphology
from hazenlib.ACRObject import ACRObject
from hazenlib.HazenTask import HazenTask
from hazenlib.logger import logger
from hazenlib.types import Measurement, Result


class ACRSlicePosition(HazenTask):
    """Slice position measurement class for DICOM images of the ACR phantom."""

    def __init__(self, **kwargs):
        if kwargs.pop("verbose", None) is not None:
            logger.warning(
                "verbose is not a supported argument for %s",
                type(self).__name__,
            )
        super().__init__(**kwargs)
        # Initialise ACR object
        self.ACR_obj = ACRObject(self.dcm_list)
        self.Y_WEDGE_OFFSET = int(20 / self.ACR_obj.dx)
        self.WEDGE_CENTER_Y = int(
            45 / self.ACR_obj.dx
        )  # Center region where it is safe to ray trace along x coordinate
        self.X_WEDGE_OFFSET = int(3 / self.ACR_obj.dx)
        self.GAUSSIAN_SIGMA = 2.5 / self.ACR_obj.dx
        self.MINIMUM_Y_PEAK_THRESHOLD = int(20 / self.ACR_obj.dx)
        self.INTERPOLATION_FACTOR = 0.2

    def run(self) -> Result:
        """Main function for performing slice position measurement
        using the first and last slices from the ACR phantom image set.

        Returns:
            dict: results are returned in a standardised dictionary structure specifying the task name, input DICOM Series Description + SeriesNumber + InstanceNumber, task measurement key-value pairs, optionally path to the generated images for visualisation
        """
        # Identify relevant slices
        dcms = [self.ACR_obj.slice_stack[0], self.ACR_obj.slice_stack[-1]]

        # Initialise results dictionary
        results = self.init_result_dict(desc=self.ACR_obj.acquisition_type())
        results.files = [self.img_desc(dcm) for dcm in dcms]

        for slice_num, dcm in zip((1, 11), dcms, strict=True):
            try:
                result = self.get_slice_position(dcm)
                results.add_measurement(
                    Measurement(
                        name="SlicePosition",
                        type="measured",
                        subtype="length difference",
                        description=f"Slice {slice_num} {self.img_desc(dcm)}",
                        value=round(result, 2),
                    ),
                )

            except Exception as e:
                logger.exception(
                    "Could not calculate the bar length difference for %s"
                    " because of : %s",
                    self.img_desc(dcm),
                    e,
                )
                traceback.print_exc(file=sys.stdout)
                continue

        # only return reports if requested
        if self.report:
            results.add_report_image(self.report_files)

        return results

    def write_report(
        self,
        dcm,
        img,
        center,
        x_pts,
        y_pts,
        interp_line_prof_L,
        interp_line_prof_R,
        interp_factor,
        pos,
        shift,
    ):
        import matplotlib.pyplot as plt

        fig, axes = plt.subplots(3, 1)
        fig.set_size_inches(8, 32)
        fig.tight_layout(pad=4)

        axes[0].imshow(img)
        axes[0].scatter(center[0], center[1], c="red")
        axes[0].plot([x_pts[0], x_pts[0]], [y_pts[0], y_pts[1]], "b")
        axes[0].plot([x_pts[1], x_pts[1]], [y_pts[0], y_pts[1]], "r")
        axes[0].axis("off")
        axes[0].set_title("Line Profiles")

        axes[1].grid()
        axes[1].plot(
            interp_factor
            * np.linspace(1, len(interp_line_prof_L), len(interp_line_prof_L))
            * self.ACR_obj.dy,
            interp_line_prof_L,
            "b",
        )
        axes[1].plot(
            interp_factor
            * np.linspace(1, len(interp_line_prof_R), len(interp_line_prof_R))
            * self.ACR_obj.dy,
            interp_line_prof_R,
            "r",
        )
        axes[1].set_title("Original Line Profiles")
        axes[1].set_xlabel("Relative Pixel Position (mm)")

        axes[2].plot(
            interp_factor
            * np.linspace(1, len(interp_line_prof_L), len(interp_line_prof_L))
            * self.ACR_obj.dy,
            interp_line_prof_L,
            "b",
        )

        shift_line = np.roll(interp_line_prof_R, pos * shift)
        if shift < 0 and pos == -1:
            shift_line[0 : np.abs(shift)] = np.nan
        elif shift < 0 and pos == 1:
            shift_line[pos * shift :] = np.nan
        elif shift > 0 and pos == -1:
            shift_line[pos * shift :] = np.nan
        else:
            shift_line[0 : np.abs(pos) * shift] = np.nan

        axes[2].grid()
        axes[2].plot(
            interp_factor
            * np.linspace(1, len(interp_line_prof_L), len(interp_line_prof_L))
            * self.ACR_obj.dy,
            shift_line,
            "r",
        )
        axes[2].set_title("Shifted Line Profiles")
        axes[2].set_xlabel("Relative Pixel Position (mm)")

        img_path = os.path.realpath(
            os.path.join(self.report_path, f"{self.img_desc(dcm)}.png")
        )
        fig.savefig(img_path)
        self.report_files.append(img_path)

        return img_path

    def find_wedges(self, img, center):
        """Find wedges in the pixel array. \n
        Investigates the top half of the phantom to locate where the wedges
        pass through the slice, and calculates the co-ordinates of these locations.

        Args:
            img (np.ndarray): dcm.pixel_array
            center (tuple[int]): dcm.pixel_array of the image mask

        Returns:
            tuple of tuples: of x and y coordinates of wedges.
        """
        # Raytrace upward to find transition zone of wedge
        yray = skimage.measure.profile_line(
            img,
            (0, center[0]),
            (center[1], center[0]),
            linewidth=int(1 / self.ACR_obj.dx),
            mode="constant",
            reduce_func=np.mean,
        ).flatten()
        abs_diff_y_profile = np.abs(np.diff(yray))
        smoothed_y_profile = scipy.ndimage.gaussian_filter1d(
            abs_diff_y_profile, self.GAUSSIAN_SIGMA
        )

        ypeaks = self.ACR_obj.find_n_highest_peaks(smoothed_y_profile, 5)
        # A properly centered phantom will not have any signal in upper region (0 < y <~20)
        # Very small hack to filter out any "peak" signal in that region.
        # This is to make sure that phantoms with a well demarcated upper edge of the wedge do not bias the line profile
        # placement towards the upper portion of the wedge box.
        ypeaks = [p for p in ypeaks[0] if p > self.MINIMUM_Y_PEAK_THRESHOLD]
        # The line profile should yield a minimum of 4 peaks.
        # The first peak denotes the start of the phantom.
        # The second peak denotes the end of the wedge region.
        # The third and fourth peaks are phantom specific since
        # they imply the edges of elements around the center
        y_center = ypeaks[1]

        # Raytrace sideways assuming static general center of wedge to detect side transitions of the wedge box.
        # The more miscentered the phantom, the more likely to get the wrong x coordinates of the wedges, but
        # the static center was selected such that other relative phantom features do not interfere unless
        # the phantom is in a really bad location.
        # Small jitters should not have as much of an impact as relative computation of this center experienced.
        # As an added benefit, we should still get a decently accurate result even if we capture the upper bound of
        # the wedges with line profiles because the delta is based on the bottom differences.
        xray = skimage.measure.profile_line(
            img,
            (self.WEDGE_CENTER_Y, 0),
            (self.WEDGE_CENTER_Y, img.shape[0]),
            linewidth=int(1 / self.ACR_obj.dx),
            mode="constant",
            reduce_func=np.mean,
        ).flatten()
        abs_diff_x_profile = np.abs(np.diff(xray))
        smoothed_x_profile = scipy.ndimage.gaussian_filter1d(
            abs_diff_x_profile, self.GAUSSIAN_SIGMA
        )

        xpeaks = self.ACR_obj.find_n_highest_peaks(smoothed_x_profile, 4)
        # The line profile should yield exactly 4 peaks unless we are off-centered.
        # The middle two peaks correspond to the edges of the wedge region on the horizontal plane.
        # The other two peaks denote the edges of the Phantom intersected on the horizontal plane.
        # A simple average of the middle 2 peaks should give the exact x center coordinate of the wedge region.
        x_center = int(np.sum(xpeaks[0][1:3]) // 2)

        logger.info("Wedge Bottom => {}, {}".format(x_center, y_center))

        return [
            int(x_center - self.X_WEDGE_OFFSET),
            int(x_center + self.X_WEDGE_OFFSET),
        ], [
            int(y_center - self.Y_WEDGE_OFFSET),
            int(y_center + self.Y_WEDGE_OFFSET),
        ]

    def get_slice_position(self, dcm):
        """Measure slice position. \n
        Locates the two opposing wedges and calculates the height difference.

        Args:
            dcm (pydicom.Dataset): DICOM image object.

        Returns:
            float: bar length difference.
        """
        logger.info(
            f"Computing slice position for slice #{dcm.InstanceNumber}"
        )
        img, rescaled, presentation = self.ACR_obj.get_presentation_pixels(dcm)
        cxy, _ = self.ACR_obj.find_phantom_center(
            rescaled, self.ACR_obj.dx, self.ACR_obj.dy
        )
        x_pts, y_pts = self.find_wedges(rescaled, cxy)
        logger.info(
            "Wedge Locations => \n{}\n{}".format(
                (x_pts[0], y_pts[0]), (x_pts[1], y_pts[1])
            )
        )

        # line profile through left wedge
        line_prof_L = skimage.measure.profile_line(
            rescaled,
            (y_pts[0], x_pts[0]),
            (y_pts[1], x_pts[0]),
            mode="constant",
        ).flatten()
        # line profile through right wedge
        line_prof_R = skimage.measure.profile_line(
            rescaled,
            (y_pts[0], x_pts[1]),
            (y_pts[1], x_pts[1]),
            mode="constant",
        ).flatten()

        # interpolation
        x = np.arange(1, len(line_prof_L) + 1)
        new_x = np.arange(
            1,
            len(line_prof_L) + self.INTERPOLATION_FACTOR,
            self.INTERPOLATION_FACTOR,
        )

        # interpolate left line profile
        interp_line_prof_L = scipy.interpolate.interp1d(x, line_prof_L)(new_x)

        # interpolate right line profile
        interp_line_prof_R = scipy.interpolate.interp1d(x, line_prof_R)(new_x)

        # difference of line profiles
        delta = interp_line_prof_L - interp_line_prof_R

        # Keep eroding delta at the edges until at least one peak is found.
        # We'll set a maximum as 20% of signal eroded as this seems like
        # a sensible upper bound for now. This may need to be adjusted
        # if evidence contradicts the above statement.
        peaks = []
        half_erosion_frac = 0.1
        for idx in range(
            max(1, int(len(delta) * half_erosion_frac)),
        ):
            # find two highest peaks
            _delta = delta[idx:-idx] if idx else delta
            peaks, _ = ACRObject.find_n_highest_peaks(
                abs(_delta),
                2,
                0.5 * np.max(abs(_delta)),
            )
            if len(peaks):
                break
        else:
            msg = "No peaks could be found"
            logger.error("%s - this test will fail", msg)
            logger.debug(
                "Parameters used for peak finding:\n"
                "\tErosion Percentage: %s\n"
                "\tMinimum peak height: %s\n" - "\tMinimum delta value: %s\n"
                "\tMaximum delta value: %s\n"
                "\tAverage delta value: %s\n",
                half_erosion_frac * 200,
                0.5 * np.max(abs(_delta)),
                np.min(abs(_delta)),
                np.max(abs(_delta)),
                np.mean(abs(_delta)),
            )
            raise ValueError(msg)
        logger.info(peaks)

        # if only one peak, set dummy range
        if len(peaks) == 1:
            # Ensure the peaks stay within a sensible range.
            peaks = [
                max(peaks[0] - 50, 0),
                min(peaks[0] + 50, len(delta) - 1),
            ]

        # set multiplier for right or left shift based on sign of peak
        pos = (
            1
            if np.max(-delta[peaks[0] : peaks[1]])
            < np.max(delta[peaks[0] : peaks[1]])
            else -1
        )

        # take line profiles in range of interest
        static_line_L = interp_line_prof_L[peaks[0] : peaks[1]]
        static_line_R = interp_line_prof_R[peaks[0] : peaks[1]]

        # create array of lag values
        lag = np.linspace(-50, 50, 101, dtype=int)

        # initialise array of errors
        err = np.zeros(len(lag))

        for k, lag_val in enumerate(lag):
            # difference of L and circularly shifted R
            difference = static_line_R - np.roll(static_line_L, lag_val)
            # set wrapped values to nan
            if lag_val > 0:
                difference[:lag_val] = np.nan
            else:
                difference[lag_val:] = np.nan

            # filler value to suppress warning when trying to calculate mean of array filled with NaN otherwise
            # calculate difference
            err[k] = (
                1e10 if np.isnan(difference).all() else np.nanmean(difference)
            )

        # find minimum non-zero error
        temp = np.argwhere(err == np.min(err[err > 0]))[0]

        # find shift corresponding to above error
        shift = -lag[temp][0] if pos == 1 else lag[temp][0]

        # calculate bar length difference
        dL = pos * np.abs(shift) * self.INTERPOLATION_FACTOR * self.ACR_obj.dy

        if self.report:
            self.write_report(
                dcm,
                rescaled,
                cxy,
                x_pts,
                y_pts,
                interp_line_prof_L,
                interp_line_prof_R,
                self.INTERPOLATION_FACTOR,
                pos,
                shift,
            )

        return dL
